from typing import Dict, Tuple

from qm.type_hinting import Number
from qm.type_hinting.config_types import DictQuaConfig, PortReferenceType, ControllerConfigType


def _prep_config(
    iq_channels: Tuple[PortReferenceType, PortReferenceType],
    adc_channels: Tuple[PortReferenceType, PortReferenceType],
    if_freq: Number,
    lo_freq: Number,
) -> DictQuaConfig:
    i_port = iq_channels[0]
    q_port = iq_channels[1]
    controller_names = set([ch[0] for ch in iq_channels] + [ch[0] for ch in adc_channels])
    controllers: Dict[str, ControllerConfigType] = {
        name: {
            "type": "opx1",
        }
        for name in controller_names
    }
    controllers[i_port[0]]["analog_outputs"] = {
        i_port[1]: {"offset": 0},
        q_port[1]: {"offset": 0},
    }
    for port in adc_channels:
        controllers[port[0]]["analog_inputs"] = {port[1]: {"offset": 0.0}}

    config: DictQuaConfig = {
        "version": 1,
        "controllers": controllers,
        "elements": {
            "to_calibrate": {
                "mixInputs": {
                    "I": i_port,
                    "Q": q_port,
                    "lo_frequency": lo_freq,
                    "mixer": "Correction_mixer",
                },
                "intermediate_frequency": if_freq,
                "operations": {},
                "digitalInputs": {},
            },
        },
        "pulses": {},
        "waveforms": {},
        "mixers": {
            "Correction_mixer": [
                {
                    "intermediate_frequency": if_freq,
                    "lo_frequency": lo_freq,
                    "correction": (1, 0, 0, 1),
                },
            ],
        },
    }

    return config



from typing import Tuple, Union, TypeVar

from qm.grpc.qua import QuaProgramRampPulse
from qm.qua._type_hinting import MessageExpressionType
from qm.qua._expressions import QuaNumberType, QuaExpressionType, QuaNumberArrayType

ChirpType = Union[
    Tuple[QuaNumberArrayType, str],
    Tuple[QuaNumberType, str],
    Tuple[QuaNumberArrayType, QuaNumberArrayType, str],
]
AmpValuesType = Tuple[
    MessageExpressionType,
    MessageExpressionType,
    MessageExpressionType,
    MessageExpressionType,
]
MeasurePulseType = Union[str, Tuple[str, AmpValuesType]]
PlayPulseType = Union[MeasurePulseType, QuaProgramRampPulse]
E = TypeVar("E")
TypeOrExpression = Union[E, QuaExpressionType]



import os
import pathlib
from typing import Union, TypeVar

import numpy

PathLike = Union[str, bytes, pathlib.Path, os.PathLike]  # type: ignore[type-arg]
Number = Union[int, float]
Value = Union[Number, bool]
NumberT = TypeVar("NumberT", int, bool, float)

NumpyNumber = Union[numpy.floating, numpy.integer]  # type: ignore[type-arg]
NumpyValue = Union[NumpyNumber, numpy.bool_]

NumpySupportedNumber = Union[Number, NumpyNumber]
NumpySupportedFloat = Union[float, numpy.floating]  # type: ignore[type-arg]
NumpySupportedValue = Union[Value, NumpyValue]



import logging

from qm.utils.deprecation_utils import throw_warning

logger = logging.getLogger("qm")
throw_warning(
    "Directly importing the logger is deprecated and will not work in future versions. "
    "To change logging levels, you can either use the function `set_logging_level(...)` from `qm.logging_utils` "
    'or you can use python native `logging.getLogger("qm").setLevel(...)`.',
    category=DeprecationWarning,
    stacklevel=2,
)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/inc_qm_api.proto
# plugin: python-betterproto
# This file has been @generated
import warnings
from dataclasses import dataclass
from typing import List

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf

from .. import (
    general_messages as _general_messages__,
    qua_config as _qua_config__,
)


class DigitalInputPortPolarity(betterproto.Enum):
    RISING = 0
    FALLING = 1


@dataclass(eq=False, repr=False)
class HighQmApiRequest(betterproto.Message):
    config: "QmConfig" = betterproto.message_field(1)
    strong_config: "_qua_config__.QuaConfig" = betterproto.message_field(2)
    quantum_machine_id: str = betterproto.string_field(3)
    set_correction: "HighQmApiRequestSetCorrection" = betterproto.message_field(
        51, group="api_oneof"
    )
    set_frequency: "HighQmApiRequestSetFrequency" = betterproto.message_field(
        52, group="api_oneof"
    )
    set_dc_offset: "HighQmApiRequestSetDcOffset" = betterproto.message_field(
        53, group="api_oneof"
    )
    set_digital_route: "HighQmApiRequestSetDigitalRoute" = betterproto.message_field(
        54, group="api_oneof"
    )
    set_io_values: "HighQmApiRequestSetIoValues" = betterproto.message_field(
        55, group="api_oneof"
    )
    set_input_dc_offset: "HighQmApiRequestSetInputDcOffset" = betterproto.message_field(
        56, group="api_oneof"
    )
    set_output_dc_offset: "HighQmApiRequestSetOutputDcOffset" = (
        betterproto.message_field(57, group="api_oneof")
    )
    set_output_filter_taps: "HighQmApiRequestSetOutputFilterTaps" = (
        betterproto.message_field(58, group="api_oneof")
    )
    set_digital_input_threshold: "HighQmApiRequestSetDigitalInputThreshold" = (
        betterproto.message_field(59, group="api_oneof")
    )
    set_digital_input_deadtime: "HighQmApiRequestSetDigitalInputDeadtime" = (
        betterproto.message_field(61, group="api_oneof")
    )
    set_digital_input_polarity: "HighQmApiRequestSetDigitalInputPolarity" = (
        betterproto.message_field(63, group="api_oneof")
    )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("config"):
            warnings.warn("HighQmApiRequest.config is deprecated", DeprecationWarning)
        if self.is_set("strong_config"):
            warnings.warn(
                "HighQmApiRequest.strong_config is deprecated", DeprecationWarning
            )
        if self.is_set("set_dc_offset"):
            warnings.warn(
                "HighQmApiRequest.set_dc_offset is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetCorrection(betterproto.Message):
    correction: "Matrix" = betterproto.message_field(1)
    qe: str = betterproto.string_field(10, group="oneof_set_correction")
    mixer: "HighQmApiRequestSetCorrectionMixerInfo" = betterproto.message_field(
        11, group="oneof_set_correction"
    )


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetCorrectionMixerInfo(betterproto.Message):
    mixer: str = betterproto.string_field(1)
    intermediate_frequency: int = betterproto.uint64_field(2)
    lo_frequency: int = betterproto.uint64_field(3)
    frequency_negative: bool = betterproto.bool_field(4)
    intermediate_frequency_double: float = betterproto.double_field(5)
    lo_frequency_double: float = betterproto.double_field(6)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("intermediate_frequency"):
            warnings.warn(
                "HighQmApiRequestSetCorrectionMixerInfo.intermediate_frequency is deprecated",
                DeprecationWarning,
            )
        if self.is_set("lo_frequency"):
            warnings.warn(
                "HighQmApiRequestSetCorrectionMixerInfo.lo_frequency is deprecated",
                DeprecationWarning,
            )


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetFrequency(betterproto.Message):
    value: float = betterproto.double_field(1)
    qe: str = betterproto.string_field(10, group="oneof_set_frequency")


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetDcOffset(betterproto.Message):
    i: float = betterproto.double_field(1)
    q: float = betterproto.double_field(2)
    qe: "HighQmApiRequestQePort" = betterproto.message_field(10, group="oneof_set_dc")


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetOutputDcOffset(betterproto.Message):
    i: float = betterproto.double_field(1)
    q: float = betterproto.double_field(2)
    qe: "HighQmApiRequestQePort" = betterproto.message_field(10, group="oneof_set_dc")


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetOutputFilterTaps(betterproto.Message):
    filter: "HighQmApiRequestAnalogOutputPortFilter" = betterproto.message_field(1)
    qe: "HighQmApiRequestQePort" = betterproto.message_field(
        10, group="oneof_set_filter_Taps"
    )


@dataclass(eq=False, repr=False)
class HighQmApiRequestAnalogOutputPortFilter(betterproto.Message):
    feedforward: List[float] = betterproto.double_field(1)
    feedback: List[float] = betterproto.double_field(2)


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetInputDcOffset(betterproto.Message):
    offset: float = betterproto.double_field(1)
    qe: "HighQmApiRequestQePort" = betterproto.message_field(10, group="oneof_set_dc")


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetDigitalRoute(betterproto.Message):
    value: int = betterproto.uint32_field(1)
    delay: "HighQmApiRequestQePort" = betterproto.message_field(
        10, group="oneof_set_digital_route"
    )
    buffer: "HighQmApiRequestQePort" = betterproto.message_field(
        11, group="oneof_set_digital_route"
    )
    tof: str = betterproto.string_field(12, group="oneof_set_digital_route")
    smearing: str = betterproto.string_field(13, group="oneof_set_digital_route")


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetIoValues(betterproto.Message):
    all: bool = betterproto.bool_field(1, group="oneof_setType")
    io_value_set_data: List[
        "HighQmApiRequestIoValueSetData"
    ] = betterproto.message_field(51)


@dataclass(eq=False, repr=False)
class HighQmApiRequestQePort(betterproto.Message):
    qe: str = betterproto.string_field(1)
    port: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class HighQmApiRequestIoValueSetData(betterproto.Message):
    io_number: int = betterproto.int32_field(1)
    int_value: int = betterproto.int32_field(2, group="oneof_IO_type")
    double_value: float = betterproto.double_field(3, group="oneof_IO_type")
    boolean_value: bool = betterproto.bool_field(4, group="oneof_IO_type")


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetDigitalInputThreshold(betterproto.Message):
    digital_port: "DigitalInputPort" = betterproto.message_field(1)
    threshold: float = betterproto.double_field(2)


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetDigitalInputPolarity(betterproto.Message):
    digital_port: "DigitalInputPort" = betterproto.message_field(1)
    polarity: "DigitalInputPortPolarity" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class HighQmApiRequestSetDigitalInputDeadtime(betterproto.Message):
    digital_port: "DigitalInputPort" = betterproto.message_field(1)
    deadtime: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class DigitalInputPort(betterproto.Message):
    controller_name: str = betterproto.string_field(1)
    port_number: int = betterproto.uint32_field(2)
    fem_number: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class HighQmApiResponse(betterproto.Message):
    ok: bool = betterproto.bool_field(1)
    errors: List["_general_messages__.ErrorMessage"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class QmConfig(betterproto.Message):
    version: int = betterproto.uint32_field(1)
    root: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class Matrix(betterproto.Message):
    v00: float = betterproto.double_field(1)
    v01: float = betterproto.double_field(2)
    v10: float = betterproto.double_field(3)
    v11: float = betterproto.double_field(4)


@dataclass(eq=False, repr=False)
class PortReference(betterproto.Message):
    controller: str = betterproto.string_field(1)
    number: int = betterproto.uint32_field(2)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/compiler.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    Dict,
    List,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf

from .. import (
    general_messages as _general_messages__,
    qm_manager as _qm_manager__,
    qua as _qua__,
)


@dataclass(eq=False, repr=False)
class QuaValues(betterproto.Message):
    int_value: int = betterproto.uint32_field(1)
    double_value: float = betterproto.double_field(2)
    boolean_value: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class CompileRequest(betterproto.Message):
    """The request message containing the user's name."""

    program: "_qua__.QuaProgram" = betterproto.message_field(1)
    job_id: str = betterproto.string_field(2)
    """used for logging inside internal services"""


@dataclass(eq=False, repr=False)
class CompileResponse(betterproto.Message):
    ok: bool = betterproto.bool_field(1)
    messages: List["CompilerMessage"] = betterproto.message_field(2)
    binary: bytes = betterproto.bytes_field(3)
    metadata: str = betterproto.string_field(4)
    debug: Dict[str, str] = betterproto.map_field(
        30, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )


@dataclass(eq=False, repr=False)
class CompilerMessage(betterproto.Message):
    message: str = betterproto.string_field(1)
    level: "_general_messages__.MessageLevel" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class ValidationResponse(betterproto.Message):
    messages: List["_qm_manager__.ConfigValidationMessage"] = betterproto.message_field(
        1
    )


@dataclass(eq=False, repr=False)
class DynamicConfig(betterproto.Message):
    """
    A config that is similar to JSON in structure and can be converted to the real config
    It has support for versioning
    """

    version: int = betterproto.uint32_field(1)
    root: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(2)



import logging
from typing import TYPE_CHECKING, Tuple, Union, Optional, Sequence

from octave_sdk.octave import RFOutput
from octave_sdk import RFOutputMode, OctaveLOSource

from qm.type_hinting import Number
from qm.api.frontend_api import FrontendApi
from qm.elements.element_inputs import MixInputs
from qm.grpc.qua_config import QuaConfigMixInputs
from qm.type_hinting.general import NumpySupportedFloat

if TYPE_CHECKING:
    from qm.octave import CalibrationDB

logger = logging.getLogger(__name__)


class UpconvertedInput(MixInputs):
    def __init__(
        self,
        name: str,
        config: QuaConfigMixInputs,
        frontend_api: FrontendApi,
        machine_id: str,
        client: RFOutput,
        port: Tuple[str, int],
        gain: Optional[float],
        calibration_db: Optional["CalibrationDB"] = None,
        use_input_attenuators: bool = False,
    ):
        super().__init__(name, config, frontend_api, machine_id)
        self._client = client
        self._lo_frequency = config.lo_frequency_double if config.lo_frequency_double else float(config.lo_frequency)
        self._port = port
        self._gain = gain
        self._calibration_db = calibration_db
        self._use_input_attenuators = use_input_attenuators

    @property
    def port(self) -> Tuple[str, int]:
        return self._port

    @property
    def lo_source(self) -> OctaveLOSource:
        return self._client.get_lo_source()

    @property
    def gain(self) -> Optional[float]:
        return self._gain

    def set_lo_frequency(self, lo_frequency: float, set_source: bool = True) -> None:
        """
        Sets the LO frequency of the synthesizer associated to element

        Args:
            lo_frequency:
            set_source:
        """
        # if self.lo_source != OctaveLOSource.Internal and self.lo_source not in self._client._port_mapping:
        #     raise SetFrequencyException(f"Cannot set frequency to an external lo source {self.lo_source.name}")

        lo_source = self.lo_source
        if set_source:
            self._client.set_lo_source(lo_source, ignore_shared_errors=True)

        self._client.set_lo_frequency(lo_source, lo_frequency)
        self.inform_element_about_lo_frequency(lo_frequency)

    def inform_element_about_lo_frequency(self, lo_frequency: float) -> None:
        self._set_config_lo_frequency(lo_frequency)
        if self._gain is not None:
            self._client.set_gain(self._gain, self.lo_frequency, self._use_input_attenuators)

    def set_use_input_attenuators(self, use_input_attenuators: bool) -> None:
        """
        Allow using the 10dB attenuators in front of the up converter IQ ports (or not).

        Args:
            use_input_attenuators: False means: never use the attenuators. True means: used it based on the desired gain.
        """
        self._use_input_attenuators = use_input_attenuators
        if self._gain is not None:
            self._client.set_gain(self._gain, self.lo_frequency, self._use_input_attenuators)
        else:
            logger.warning("Gain is not set, so the input attenuators cannot be set.")

    def set_lo_source(self, lo_port: OctaveLOSource) -> None:
        """
        Sets the source of LO going to the upconverter associated with element.
        Args:
            lo_port:
        """
        self._client.set_lo_source(lo_port, ignore_shared_errors=True)

    def set_rf_output_mode(self, switch_mode: RFOutputMode) -> None:
        """
        Configures the output switch of the upconverter associated to element.
        switch_mode can be either: 'always_on', 'always_off', 'normal' or 'inverted'
        When in 'normal' mode a high trigger will turn the switch on and a low trigger will turn it off
        When in 'inverted' mode a high trigger will turn the switch off and a low trigger will turn it on
        When in 'always_on' the switch will be permanently on. When in 'always_off' mode the switch will
        be permanently off.

        Args:
            switch_mode:
        """
        self._client.set_output(switch_mode)

    def set_rf_output_gain(self, gain_in_db: float) -> None:
        """
        Sets the RF output gain for the up-converter associated with the element.
        RF_gain is in steps of 0.5dB from -20 to +24 and is referring
        to the maximum OPX output power of 4dBm (=0.5V pk-pk) So for a value of -24
        for example, an IF signal coming from the OPX at max power (4dBm) will be
        upconverted and come out of Octave at -20dBm

        Args:
            gain_in_db:
        """
        # a. use value custom set in qm.octave.update_external
        # b. use value from config
        if not -20 <= gain_in_db <= 20:
            raise ValueError(f"Gain must be between -20 and 20 dB, got {gain_in_db}")
        self._client.set_gain(gain_in_db, self.lo_frequency, self._use_input_attenuators)
        self._gain = gain_in_db


class UpconvertedInputNewApi(UpconvertedInput):
    def __init__(
        self,
        name: str,
        config: QuaConfigMixInputs,
        client: RFOutput,
        port: Tuple[str, int],
        gain: Optional[float],
        calibration_db: Optional["CalibrationDB"] = None,
        use_input_attenuators: bool = False,
    ):
        super().__init__(
            name,
            config,
            frontend_api=None,  # type: ignore[arg-type]
            machine_id=None,  # type: ignore[arg-type]
            client=client,
            port=port,
            gain=gain,
            calibration_db=calibration_db,
            use_input_attenuators=use_input_attenuators,
        )

    def set_output_dc_offset(self, i_offset: Optional[float] = None, q_offset: Optional[float] = None) -> None:
        raise NotImplementedError

    def set_output_filter(
        self,
        input_name: str,
        feedforward: Union[Sequence[NumpySupportedFloat], None],
        feedback: Union[Sequence[NumpySupportedFloat], None],
    ) -> None:
        raise NotImplementedError

    def set_mixer_correction(
        self,
        intermediate_frequency: Number,
        lo_frequency: Number,
        values: Tuple[float, float, float, float],
    ) -> None:
        raise NotImplementedError



from typing import Tuple

import numpy as np

Correction = Tuple[float, float, float, float]


def convert_to_correction(gain: float, phase: float) -> Correction:
    """
    Convert gain and phase to a correction matrix.
    """
    s = phase
    c = np.polyval([-3.125, 1.5, 1], s**2)
    g_plus = np.polyval([0.5, 1, 1], gain)
    g_minus = np.polyval([0.5, -1, 1], gain)

    c00 = float(g_plus * c)
    c01 = float(g_plus * s)
    c10 = float(g_minus * s)
    c11 = float(g_minus * c)

    return c00, c01, c10, c11



from qm.type_hinting.config_types import DictQuaConfig
from qm.type_hinting.general import Value, Number, NumberT, PathLike, NumpySupportedValue, NumpySupportedNumber

__all__ = ["DictQuaConfig", "Number", "Value", "PathLike", "NumberT", "NumpySupportedValue", "NumpySupportedNumber"]



import time
import logging
from typing import Any, Dict, Generic, TypeVar, Callable

SERVICE_HEADER_NAME = "x-grpc-service"

T = TypeVar("T")


def create_input_stream_name(name: str) -> str:
    return f"input_stream_{name}"


def run_until_with_timeout(
    on_iteration_callback: Callable[[], bool],
    # The type ignore can be removed in 3.12 with `T = TypeVar("T", default=None)`
    on_complete_callback: Callable[[], T] = lambda: None,  # type: ignore[assignment, return-value]
    timeout: float = float("infinity"),
    loop_interval: float = 0.1,
    timeout_message: str = "Timeout Exceeded",
) -> T:
    """

    Args:
        on_iteration_callback: A callback that returns bool that is called every loop iteration.
            If True is returned, the loop is complete and on_complete_callback is called.

        on_complete_callback: A callback that is called when the loop is completed.
            This function returns the return value of on_complete_callback

        timeout: The timeout in seconds for on_iteration_callback to return True,
            on Zero the loop is executed once.

        loop_interval: The interval (in seconds) between each loop execution.

        timeout_message: The message of the TimeoutError exception.
    """
    if timeout < 0:
        raise ValueError("timeout cannot be smaller than 0")

    start = time.time()
    end = start + timeout

    while True:
        if on_iteration_callback():
            return on_complete_callback()

        time.sleep(loop_interval)

        if time.time() >= end:
            raise TimeoutError(timeout_message)


_T = TypeVar("_T")


class Singleton(type, Generic[_T]):
    _instances: Dict["Singleton[_T]", _T] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> _T:
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


def is_debug() -> bool:
    return logging.getLogger("qm").level <= logging.DEBUG



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/inc_qua_config.proto
# plugin: python-betterproto
# This file has been @generated
import warnings
from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf


class QuaConfigVoltageLevel(betterproto.Enum):
    LVTTL = 0
    TTL = 1


class QuaConfigOutputSwitchState(betterproto.Enum):
    unset = 0
    always_on = 1
    always_off = 2
    triggered = 3
    triggered_reversed = 4


class QuaConfigOctoDacAnalogOutputPortDecOutputMode(betterproto.Enum):
    direct = 0
    amplified = 1


class QuaConfigOctoDacAnalogOutputPortDecSamplingRate(betterproto.Enum):
    Undefined = 0
    GSPS1 = 1
    GSPS2 = 2


class QuaConfigOctoDacAnalogOutputPortDecSamplingRateMode(betterproto.Enum):
    unset = 0
    mw = 1
    pulse = 2


class QuaConfigDigitalInputPortDecPolarity(betterproto.Enum):
    RISING = 0
    FALLING = 1


class QuaConfigOutputPulseParametersPolarity(betterproto.Enum):
    ASCENDING = 0
    DESCENDING = 1


class QuaConfigPulseDecOperation(betterproto.Enum):
    MEASUREMENT = 0
    CONTROL = 1


class QuaConfigOctaveSynthesizerOutputName(betterproto.Enum):
    synth1 = 0
    synth2 = 1
    synth3 = 2
    synth4 = 3
    synth5 = 4


class QuaConfigOctaveLoSourceInput(betterproto.Enum):
    not_set = 0
    internal = 1
    external = 2
    analyzer = 3


class QuaConfigOctaveLoopbackInput(betterproto.Enum):
    undefined = 0
    LO1 = 1
    LO2 = 2
    LO3 = 3
    LO4 = 4
    LO5 = 5
    Dmd1LO = 6
    Dmd2LO = 7


class QuaConfigOctaveDownconverterRfSource(betterproto.Enum):
    rf_in = 0
    loopback_1 = 1
    loopback_2 = 2
    loopback_3 = 3
    loopback_4 = 4
    loopback_5 = 5


class QuaConfigOctaveOutputSwitchState(betterproto.Enum):
    unset = 0
    always_on = 1
    always_off = 2
    triggered = 3
    triggered_reversed = 4


class QuaConfigOctaveIfMode(betterproto.Enum):
    direct = 0
    mixer = 1
    envelope = 2
    off = 3


@dataclass(eq=False, repr=False)
class QuaConfig(betterproto.Message):
    v1_beta: "QuaConfigQuaConfigV1" = betterproto.message_field(
        1, group="config_version"
    )
    """beta suffix means the version is not fixed yet and can changed"""

    revision: int = betterproto.int32_field(11)


@dataclass(eq=False, repr=False)
class QuaConfigQuaConfigV1(betterproto.Message):
    controllers: Dict[str, "QuaConfigControllerDec"] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    control_devices: Dict[str, "QuaConfigDeviceDec"] = betterproto.map_field(
        20, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    oscillators: Dict[str, "QuaConfigOscillator"] = betterproto.map_field(
        9, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    elements: Dict[str, "QuaConfigElementDec"] = betterproto.map_field(
        3, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    pulses: Dict[str, "QuaConfigPulseDec"] = betterproto.map_field(
        4, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    mixers: Dict[str, "QuaConfigMixerDec"] = betterproto.map_field(
        5, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    waveforms: Dict[str, "QuaConfigWaveformDec"] = betterproto.map_field(
        6, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    digital_waveforms: Dict[str, "QuaConfigDigitalWaveformDec"] = betterproto.map_field(
        7, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    integration_weights: Dict[
        str, "QuaConfigIntegrationWeightDec"
    ] = betterproto.map_field(8, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE)
    octaves: Dict[str, "QuaConfigOctaveConfig"] = betterproto.map_field(
        11, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("controllers"):
            warnings.warn(
                "QuaConfigQuaConfigV1.controllers is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaConfigDeviceDec(betterproto.Message):
    fems: Dict[int, "QuaConfigFemTypes"] = betterproto.map_field(
        1, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )


@dataclass(eq=False, repr=False)
class QuaConfigFemTypes(betterproto.Message):
    opx: "QuaConfigControllerDec" = betterproto.message_field(
        1, group="fem_type_one_of"
    )
    octo_dac: "QuaConfigOctoDacFemDec" = betterproto.message_field(
        2, group="fem_type_one_of"
    )
    microwave: "QuaConfigMicrowaveFemDec" = betterproto.message_field(
        3, group="fem_type_one_of"
    )


@dataclass(eq=False, repr=False)
class QuaConfigControllerDec(betterproto.Message):
    type: str = betterproto.string_field(1)
    analog_outputs: Dict[int, "QuaConfigAnalogOutputPortDec"] = betterproto.map_field(
        2, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )
    analog_inputs: Dict[int, "QuaConfigAnalogInputPortDec"] = betterproto.map_field(
        3, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )
    digital_outputs: Dict[int, "QuaConfigDigitalOutputPortDec"] = betterproto.map_field(
        4, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )
    digital_inputs: Dict[int, "QuaConfigDigitalInputPortDec"] = betterproto.map_field(
        5, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )


@dataclass(eq=False, repr=False)
class QuaConfigOctoDacFemDec(betterproto.Message):
    analog_outputs: Dict[
        int, "QuaConfigOctoDacAnalogOutputPortDec"
    ] = betterproto.map_field(2, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE)
    analog_inputs: Dict[int, "QuaConfigAnalogInputPortDec"] = betterproto.map_field(
        3, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )
    digital_outputs: Dict[int, "QuaConfigDigitalOutputPortDec"] = betterproto.map_field(
        4, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )
    digital_inputs: Dict[int, "QuaConfigDigitalInputPortDec"] = betterproto.map_field(
        5, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )


@dataclass(eq=False, repr=False)
class QuaConfigMicrowaveFemDec(betterproto.Message):
    analog_outputs: Dict[
        int, "QuaConfigMicrowaveAnalogOutputPortDec"
    ] = betterproto.map_field(2, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE)
    analog_inputs: Dict[
        int, "QuaConfigMicrowaveAnalogInputPortDec"
    ] = betterproto.map_field(3, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE)
    digital_outputs: Dict[int, "QuaConfigDigitalOutputPortDec"] = betterproto.map_field(
        4, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )
    digital_inputs: Dict[int, "QuaConfigDigitalInputPortDec"] = betterproto.map_field(
        5, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )


@dataclass(eq=False, repr=False)
class QuaConfigAnalogOutputPortDec(betterproto.Message):
    offset: float = betterproto.double_field(1)
    filter: "QuaConfigAnalogOutputPortFilter" = betterproto.message_field(2)
    delay: int = betterproto.uint32_field(3)
    channel_weights: Dict[int, float] = betterproto.map_field(
        4, betterproto.TYPE_UINT32, betterproto.TYPE_DOUBLE
    )
    shareable: bool = betterproto.bool_field(5)
    crosstalk: Dict[int, float] = betterproto.map_field(
        6, betterproto.TYPE_UINT32, betterproto.TYPE_DOUBLE
    )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("channel_weights"):
            warnings.warn(
                "QuaConfigAnalogOutputPortDec.channel_weights is deprecated",
                DeprecationWarning,
            )


@dataclass(eq=False, repr=False)
class QuaConfigMicrowaveAnalogOutputPortDec(betterproto.Message):
    sampling_rate: float = betterproto.double_field(1)
    full_scale_power_dbm: int = betterproto.int32_field(2)
    band: int = betterproto.int32_field(3)
    delay: int = betterproto.uint32_field(4)
    shareable: bool = betterproto.bool_field(5)
    upconverters: Dict[int, "QuaConfigUpConverterConfigDec"] = betterproto.map_field(
        6, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )


@dataclass(eq=False, repr=False)
class QuaConfigUpConverterConfigDec(betterproto.Message):
    frequency: float = betterproto.double_field(1)


@dataclass(eq=False, repr=False)
class QuaConfigMicrowaveAnalogInputPortDec(betterproto.Message):
    sampling_rate: float = betterproto.double_field(1)
    gain_db: int = betterproto.int32_field(2)
    shareable: bool = betterproto.bool_field(3)
    band: int = betterproto.int32_field(4)
    downconverter: "QuaConfigDownConverterConfigDec" = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class QuaConfigDownConverterConfigDec(betterproto.Message):
    frequency: float = betterproto.double_field(1)


@dataclass(eq=False, repr=False)
class QuaConfigOctoDacAnalogOutputPortDec(betterproto.Message):
    offset: float = betterproto.double_field(1)
    filter: "QuaConfigAnalogOutputPortFilter" = betterproto.message_field(2)
    delay: int = betterproto.uint32_field(3)
    shareable: bool = betterproto.bool_field(5)
    crosstalk: Dict[int, float] = betterproto.map_field(
        6, betterproto.TYPE_UINT32, betterproto.TYPE_DOUBLE
    )
    sampling_rate: "QuaConfigOctoDacAnalogOutputPortDecSamplingRate" = (
        betterproto.enum_field(7)
    )
    upsampling_mode: "QuaConfigOctoDacAnalogOutputPortDecSamplingRateMode" = (
        betterproto.enum_field(8)
    )
    output_mode: "QuaConfigOctoDacAnalogOutputPortDecOutputMode" = (
        betterproto.enum_field(9)
    )


@dataclass(eq=False, repr=False)
class QuaConfigAnalogOutputPortFilter(betterproto.Message):
    feedforward: List[float] = betterproto.double_field(1)
    feedback: List[float] = betterproto.double_field(2)


@dataclass(eq=False, repr=False)
class QuaConfigAnalogInputPortDec(betterproto.Message):
    offset: float = betterproto.double_field(1)
    gain_db: Optional[int] = betterproto.message_field(2, wraps=betterproto.TYPE_INT32)
    shareable: bool = betterproto.bool_field(3)
    sampling_rate: Optional[float] = betterproto.double_field(4, optional=True)


@dataclass(eq=False, repr=False)
class QuaConfigDigitalOutputPortDec(betterproto.Message):
    shareable: bool = betterproto.bool_field(1)
    inverted: bool = betterproto.bool_field(2)
    level: Optional["QuaConfigVoltageLevel"] = betterproto.enum_field(3, optional=True)


@dataclass(eq=False, repr=False)
class QuaConfigDigitalInputPortDec(betterproto.Message):
    deadtime: int = betterproto.uint32_field(1)
    polarity: "QuaConfigDigitalInputPortDecPolarity" = betterproto.enum_field(2)
    threshold: float = betterproto.double_field(3)
    shareable: bool = betterproto.bool_field(4)
    level: Optional["QuaConfigVoltageLevel"] = betterproto.enum_field(5, optional=True)


@dataclass(eq=False, repr=False)
class QuaConfigMixerRef(betterproto.Message):
    mixer: str = betterproto.string_field(1)
    lo_frequency: int = betterproto.uint64_field(2)
    lo_frequency_double: float = betterproto.double_field(3)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("lo_frequency"):
            warnings.warn(
                "QuaConfigMixerRef.lo_frequency is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaConfigOscillator(betterproto.Message):
    intermediate_frequency: Optional[int] = betterproto.message_field(
        1, wraps=betterproto.TYPE_INT64
    )
    mixer: "QuaConfigMixerRef" = betterproto.message_field(2)
    intermediate_frequency_double: float = betterproto.double_field(3)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("intermediate_frequency"):
            warnings.warn(
                "QuaConfigOscillator.intermediate_frequency is deprecated",
                DeprecationWarning,
            )


@dataclass(eq=False, repr=False)
class QuaConfigSingleInput(betterproto.Message):
    port: "QuaConfigDacPortReference" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class QuaConfigMixInputs(betterproto.Message):
    i: "QuaConfigDacPortReference" = betterproto.message_field(1)
    q: "QuaConfigDacPortReference" = betterproto.message_field(2)
    mixer: str = betterproto.string_field(3)
    lo_frequency: int = betterproto.uint64_field(4)
    lo_frequency_double: float = betterproto.double_field(5)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("lo_frequency"):
            warnings.warn(
                "QuaConfigMixInputs.lo_frequency is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaConfigSingleInputCollection(betterproto.Message):
    inputs: Dict[str, "QuaConfigDacPortReference"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )


@dataclass(eq=False, repr=False)
class QuaConfigMultipleInputs(betterproto.Message):
    inputs: Dict[str, "QuaConfigDacPortReference"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )


@dataclass(eq=False, repr=False)
class QuaConfigGeneralPortReference(betterproto.Message):
    device_name: str = betterproto.string_field(1)
    port: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class QuaConfigMultipleOutputs(betterproto.Message):
    port_references: Dict[str, "QuaConfigAdcPortReference"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )


@dataclass(eq=False, repr=False)
class QuaConfigElementDec(betterproto.Message):
    outputs: Dict[str, "QuaConfigAdcPortReference"] = betterproto.map_field(
        3, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    digital_inputs: Dict[
        str, "QuaConfigDigitalInputPortReference"
    ] = betterproto.map_field(4, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE)
    digital_outputs: Dict[
        str, "QuaConfigDigitalOutputPortReference"
    ] = betterproto.map_field(9, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE)
    rf_inputs: Dict[str, "QuaConfigGeneralPortReference"] = betterproto.map_field(
        10, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    rf_outputs: Dict[str, "QuaConfigGeneralPortReference"] = betterproto.map_field(
        11, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    single_input: "QuaConfigSingleInput" = betterproto.message_field(
        5, group="element_inputs_one_of"
    )
    mix_inputs: "QuaConfigMixInputs" = betterproto.message_field(
        6, group="element_inputs_one_of"
    )
    single_input_collection: "QuaConfigSingleInputCollection" = (
        betterproto.message_field(50, group="element_inputs_one_of")
    )
    multiple_inputs: "QuaConfigMultipleInputs" = betterproto.message_field(
        51, group="element_inputs_one_of"
    )
    microwave_input: "QuaConfigMicrowaveInputPortReference" = betterproto.message_field(
        52, group="element_inputs_one_of"
    )
    microwave_output: "QuaConfigMicrowaveOutputPortReference" = (
        betterproto.message_field(53, group="element_outputs_one_of")
    )
    multiple_outputs: "QuaConfigMultipleOutputs" = betterproto.message_field(
        54, group="element_outputs_one_of"
    )
    time_of_flight: Optional[int] = betterproto.message_field(
        7, wraps=betterproto.TYPE_UINT32
    )
    smearing: Optional[int] = betterproto.message_field(
        8, wraps=betterproto.TYPE_UINT32
    )
    intermediate_frequency: Optional[int] = betterproto.message_field(
        20, wraps=betterproto.TYPE_UINT64
    )
    intermediate_frequency_double: float = betterproto.double_field(24)
    intermediate_frequency_negative: bool = betterproto.bool_field(23)
    operations: Dict[str, str] = betterproto.map_field(
        21, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )
    measurement_qe: Optional[str] = betterproto.message_field(
        22, wraps=betterproto.TYPE_STRING
    )
    output_pulse_parameters: "QuaConfigOutputPulseParameters" = (
        betterproto.message_field(30)
    )
    hold_offset: "QuaConfigHoldOffset" = betterproto.message_field(31)
    sticky: "QuaConfigSticky" = betterproto.message_field(32)
    thread: "QuaConfigElementThread" = betterproto.message_field(40)
    intermediate_frequency_oscillator: Optional[int] = betterproto.message_field(
        61, wraps=betterproto.TYPE_INT64, group="oscillator_one_of"
    )
    intermediate_frequency_oscillator_double: float = betterproto.double_field(
        64, group="oscillator_one_of"
    )
    named_oscillator: Optional[str] = betterproto.message_field(
        62, wraps=betterproto.TYPE_STRING, group="oscillator_one_of"
    )
    no_oscillator: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        63, group="oscillator_one_of"
    )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("outputs"):
            warnings.warn(
                "QuaConfigElementDec.outputs is deprecated", DeprecationWarning
            )
        if self.is_set("intermediate_frequency"):
            warnings.warn(
                "QuaConfigElementDec.intermediate_frequency is deprecated",
                DeprecationWarning,
            )
        if self.is_set("hold_offset"):
            warnings.warn(
                "QuaConfigElementDec.hold_offset is deprecated", DeprecationWarning
            )
        if self.is_set("intermediate_frequency_oscillator"):
            warnings.warn(
                "QuaConfigElementDec.intermediate_frequency_oscillator is deprecated",
                DeprecationWarning,
            )
        if self.is_set("intermediate_frequency_oscillator_double"):
            warnings.warn(
                "QuaConfigElementDec.intermediate_frequency_oscillator_double is deprecated",
                DeprecationWarning,
            )


@dataclass(eq=False, repr=False)
class QuaConfigElementThread(betterproto.Message):
    thread_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class QuaConfigOutputPulseParameters(betterproto.Message):
    threshold: int = betterproto.uint32_field(1)
    table: List[int] = betterproto.uint32_field(2)
    signal_threshold: int = betterproto.int32_field(3)
    signal_polarity: "QuaConfigOutputPulseParametersPolarity" = betterproto.enum_field(
        4
    )
    derivative_threshold: int = betterproto.int32_field(5)
    derivative_polarity: "QuaConfigOutputPulseParametersPolarity" = (
        betterproto.enum_field(6)
    )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("threshold"):
            warnings.warn(
                "QuaConfigOutputPulseParameters.threshold is deprecated",
                DeprecationWarning,
            )
        if self.is_set("table"):
            warnings.warn(
                "QuaConfigOutputPulseParameters.table is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaConfigHoldOffset(betterproto.Message):
    duration: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class QuaConfigSticky(betterproto.Message):
    analog: bool = betterproto.bool_field(1)
    digital: bool = betterproto.bool_field(2)
    duration: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class QuaConfigDacPortReference(betterproto.Message):
    controller: str = betterproto.string_field(1)
    number: int = betterproto.uint32_field(2)
    fem: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class QuaConfigAdcPortReference(betterproto.Message):
    controller: str = betterproto.string_field(1)
    number: int = betterproto.uint32_field(2)
    fem: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class QuaConfigDigitalInputPortReference(betterproto.Message):
    port: "QuaConfigPortReference" = betterproto.message_field(1)
    delay: int = betterproto.uint32_field(2)
    buffer: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class QuaConfigDigitalOutputPortReference(betterproto.Message):
    port: "QuaConfigPortReference" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class QuaConfigPortReference(betterproto.Message):
    controller: str = betterproto.string_field(1)
    number: int = betterproto.uint32_field(2)
    fem: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class QuaConfigMicrowaveInputPortReference(betterproto.Message):
    port: "QuaConfigDacPortReference" = betterproto.message_field(1)
    upconverter: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class QuaConfigMicrowaveOutputPortReference(betterproto.Message):
    port: "QuaConfigAdcPortReference" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class QuaConfigPulseDec(betterproto.Message):
    length: int = betterproto.uint32_field(2)
    operation: int = betterproto.uint32_field(3)
    waveforms: Dict[str, str] = betterproto.map_field(
        4, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )
    digital_marker: Optional[str] = betterproto.message_field(
        5, wraps=betterproto.TYPE_STRING
    )
    integration_weights: Dict[str, str] = betterproto.map_field(
        8, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )


@dataclass(eq=False, repr=False)
class QuaConfigWaveformDec(betterproto.Message):
    arbitrary: "QuaConfigArbitraryWaveformDec" = betterproto.message_field(
        1, group="waveform_oneof"
    )
    constant: "QuaConfigConstantWaveformDec" = betterproto.message_field(
        2, group="waveform_oneof"
    )
    compressed: "QuaConfigCompressedWaveformDec" = betterproto.message_field(
        3, group="waveform_oneof"
    )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("compressed"):
            warnings.warn(
                "QuaConfigWaveformDec.compressed is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class QuaConfigArbitraryWaveformDec(betterproto.Message):
    samples: List[float] = betterproto.double_field(1)
    multiplier: float = betterproto.double_field(2)
    deprecated_max_allowed_error: float = betterproto.double_field(3)
    max_allowed_error: Optional[float] = betterproto.message_field(
        4, wraps=betterproto.TYPE_DOUBLE
    )
    sampling_rate: Optional[float] = betterproto.message_field(
        5, wraps=betterproto.TYPE_DOUBLE
    )
    is_overridable: bool = betterproto.bool_field(6)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("deprecated_max_allowed_error"):
            warnings.warn(
                "QuaConfigArbitraryWaveformDec.deprecated_max_allowed_error is deprecated",
                DeprecationWarning,
            )


@dataclass(eq=False, repr=False)
class QuaConfigCompressedWaveformDec(betterproto.Message):
    samples: List[float] = betterproto.double_field(1)
    sample_rate: float = betterproto.double_field(2)


@dataclass(eq=False, repr=False)
class QuaConfigConstantWaveformDec(betterproto.Message):
    sample: float = betterproto.double_field(1)
    multiplier: float = betterproto.double_field(2)


@dataclass(eq=False, repr=False)
class QuaConfigDigitalWaveformDec(betterproto.Message):
    samples: List["QuaConfigDigitalWaveformSample"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class QuaConfigDigitalWaveformSample(betterproto.Message):
    value: bool = betterproto.bool_field(1)
    length: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class QuaConfigMixerDec(betterproto.Message):
    correction: List["QuaConfigCorrectionEntry"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class QuaConfigIntegrationWeightDec(betterproto.Message):
    cosine: List["QuaConfigIntegrationWeightSample"] = betterproto.message_field(1)
    sine: List["QuaConfigIntegrationWeightSample"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class QuaConfigIntegrationWeightSample(betterproto.Message):
    value: float = betterproto.double_field(1)
    length: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class QuaConfigCorrectionEntry(betterproto.Message):
    frequency: int = betterproto.uint64_field(1)
    lo_frequency: int = betterproto.uint64_field(2)
    correction: "QuaConfigMatrix" = betterproto.message_field(3)
    frequency_negative: bool = betterproto.bool_field(4)
    frequency_double: float = betterproto.double_field(5)
    lo_frequency_double: float = betterproto.double_field(6)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("frequency"):
            warnings.warn(
                "QuaConfigCorrectionEntry.frequency is deprecated", DeprecationWarning
            )
        if self.is_set("lo_frequency"):
            warnings.warn(
                "QuaConfigCorrectionEntry.lo_frequency is deprecated",
                DeprecationWarning,
            )


@dataclass(eq=False, repr=False)
class QuaConfigMatrix(betterproto.Message):
    v00: float = betterproto.double_field(1)
    v01: float = betterproto.double_field(2)
    v10: float = betterproto.double_field(3)
    v11: float = betterproto.double_field(4)


@dataclass(eq=False, repr=False)
class QuaConfigOctave(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class QuaConfigOctaveConfig(betterproto.Message):
    loopbacks: List["QuaConfigOctaveLoopback"] = betterproto.message_field(1)
    rf_outputs: Dict[int, "QuaConfigOctaveRfOutputConfig"] = betterproto.map_field(
        2, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )
    rf_inputs: Dict[int, "QuaConfigOctaveRfInputConfig"] = betterproto.map_field(
        3, betterproto.TYPE_UINT32, betterproto.TYPE_MESSAGE
    )
    if_outputs: "QuaConfigOctaveIfOutputsConfig" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class QuaConfigOctaveLoopback(betterproto.Message):
    lo_source_input: "QuaConfigOctaveLoopbackInput" = betterproto.enum_field(1)
    lo_source_generator: "QuaConfigOctaveSynthesizerPort" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class QuaConfigOctaveSynthesizerPort(betterproto.Message):
    device_name: str = betterproto.string_field(1)
    port_name: "QuaConfigOctaveSynthesizerOutputName" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class QuaConfigOctaveRfOutputConfig(betterproto.Message):
    lo_frequency: float = betterproto.double_field(1)
    lo_source: "QuaConfigOctaveLoSourceInput" = betterproto.enum_field(2)
    output_mode: "QuaConfigOctaveOutputSwitchState" = betterproto.enum_field(3)
    gain: float = betterproto.float_field(4)
    input_attenuators: bool = betterproto.bool_field(5)
    i_connection: "QuaConfigDacPortReference" = betterproto.message_field(6)
    q_connection: "QuaConfigDacPortReference" = betterproto.message_field(7)


@dataclass(eq=False, repr=False)
class QuaConfigOctaveRfInputConfig(betterproto.Message):
    rf_source: "QuaConfigOctaveDownconverterRfSource" = betterproto.enum_field(1)
    lo_frequency: float = betterproto.double_field(2)
    lo_source: "QuaConfigOctaveLoSourceInput" = betterproto.enum_field(3)
    if_mode_i: "QuaConfigOctaveIfMode" = betterproto.enum_field(4)
    if_mode_q: "QuaConfigOctaveIfMode" = betterproto.enum_field(5)


@dataclass(eq=False, repr=False)
class QuaConfigOctaveSingleIfOutputConfig(betterproto.Message):
    port: "QuaConfigAdcPortReference" = betterproto.message_field(1)
    name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class QuaConfigOctaveIfOutputsConfig(betterproto.Message):
    if_out1: "QuaConfigOctaveSingleIfOutputConfig" = betterproto.message_field(1)
    if_out2: "QuaConfigOctaveSingleIfOutputConfig" = betterproto.message_field(2)



from qm.utils.deprecation_utils import throw_warning
from qm.quantum_machine import QuantumMachine  # noqa

throw_warning(
    "'qm.QuantumMachine.QuantumMachine' is moved as of 1.2.0 and will be removed in 1.4.0. "
    "use 'qm.QuantumMachine' instead",
    category=DeprecationWarning,
    stacklevel=2,
)



from typing import Dict, Union, Optional, overload

import betterproto

from qm.api.frontend_api import FrontendApi
from qm._octaves_container import OctavesContainer
from qm.octave.octave_config import QmOctaveConfig
from qm.elements.element import Element, AllElements, NewApiUpconvertedElement
from qm.elements.element_outputs import NoOutput, ElementOutput, DownconvertedOutput
from qm.elements.element_inputs import (
    NoInput,
    MixInputs,
    SingleInput,
    MicrowaveInput,
    MultipleInputs,
    ElementInputGRPCType,
    SingleInputCollection,
)
from qm.grpc.qua_config import (
    QuaConfig,
    QuaConfigMixInputs,
    QuaConfigElementDec,
    QuaConfigSingleInput,
    QuaConfigMultipleInputs,
    QuaConfigGeneralPortReference,
    QuaConfigSingleInputCollection,
    QuaConfigMicrowaveInputPortReference,
)


class ElementNotFound(KeyError):
    def __init__(self, key: str):
        self._key = key

    def __str__(self) -> str:
        return f"Element with the key {self._key} was not found."


class UnknownElementType(ValueError):
    pass


class ElementsDB(Dict[str, AllElements]):
    def __missing__(self, key: str) -> None:
        raise ElementNotFound(key)


def init_elements(
    pb_config: QuaConfig,
    frontend_api: FrontendApi,
    machine_id: str,
    octave_config: Optional[QmOctaveConfig] = None,
) -> ElementsDB:
    elements = {}
    _octave_container = OctavesContainer(pb_config, octave_config)
    for name, element_config in pb_config.v1_beta.elements.items():
        _, element_inputs = betterproto.which_one_of(element_config, "element_inputs_one_of")
        input_inst = _get_element_input(
            element_config, element_inputs, name, frontend_api, machine_id, _octave_container
        )
        rf_output = _get_element_rf_output(element_config.rf_outputs, _octave_container)

        elements[name] = Element(
            name=name,
            config=element_config,
            api=frontend_api,
            machine_id=machine_id,
            element_input=input_inst,
            element_output=rf_output,
        )
    return ElementsDB(elements)


@overload
def _get_element_input(
    element_config: QuaConfigElementDec,
    element_inputs: None,
    name: str,
    frontend_api: FrontendApi,
    machine_id: str,
    octave_container: OctavesContainer,
) -> NoInput:
    pass


@overload
def _get_element_input(
    element_config: QuaConfigElementDec,
    element_inputs: QuaConfigMixInputs,
    name: str,
    frontend_api: FrontendApi,
    machine_id: str,
    octave_container: OctavesContainer,
) -> MixInputs:
    pass


@overload
def _get_element_input(
    element_config: QuaConfigElementDec,
    element_inputs: QuaConfigSingleInput,
    name: str,
    frontend_api: FrontendApi,
    machine_id: str,
    octave_container: OctavesContainer,
) -> SingleInput:
    pass


@overload
def _get_element_input(
    element_config: QuaConfigElementDec,
    element_inputs: QuaConfigMultipleInputs,
    name: str,
    frontend_api: FrontendApi,
    machine_id: str,
    octave_container: OctavesContainer,
) -> MultipleInputs:
    pass


@overload
def _get_element_input(
    element_config: QuaConfigElementDec,
    element_inputs: QuaConfigSingleInputCollection,
    name: str,
    frontend_api: FrontendApi,
    machine_id: str,
    octave_container: OctavesContainer,
) -> SingleInputCollection:
    pass


@overload
def _get_element_input(
    element_config: QuaConfigElementDec,
    element_inputs: QuaConfigMicrowaveInputPortReference,
    name: str,
    frontend_api: FrontendApi,
    machine_id: str,
    octave_container: OctavesContainer,
) -> MicrowaveInput:
    pass


def _get_element_input(
    element_config: QuaConfigElementDec,
    element_inputs: ElementInputGRPCType,
    name: str,
    frontend_api: FrontendApi,
    machine_id: str,
    octave_container: OctavesContainer,
) -> Union[NoInput, MixInputs, SingleInput, MultipleInputs, SingleInputCollection, MicrowaveInput]:
    if element_inputs is None:
        return NoInput(name, element_inputs, frontend_api, machine_id)
    if isinstance(element_inputs, QuaConfigMixInputs):
        return octave_container.create_mix_inputs(element_config, name, frontend_api, machine_id)
    if isinstance(element_inputs, QuaConfigSingleInput):
        return SingleInput(name, element_inputs, frontend_api, machine_id)
    if isinstance(element_inputs, QuaConfigMultipleInputs):
        return MultipleInputs(name, element_inputs, frontend_api, machine_id)
    if isinstance(element_inputs, QuaConfigSingleInputCollection):
        return SingleInputCollection(name, element_inputs, frontend_api, machine_id)
    if isinstance(element_inputs, QuaConfigMicrowaveInputPortReference):
        return MicrowaveInput(name, element_inputs, frontend_api, machine_id)
    raise UnknownElementType(f"Element {name} is of unknown type - {type(element_inputs)}.")


def _get_element_rf_output(
    rf_outputs: Dict[str, QuaConfigGeneralPortReference], octave_container: OctavesContainer
) -> ElementOutput:
    downconverter = octave_container.get_downconverter(rf_outputs)
    if downconverter is not None:  # I prefer isinstace, but this allows for easier testing
        return DownconvertedOutput(downconverter)
    return NoOutput()


class UpconvertedElementsDB(Dict[str, NewApiUpconvertedElement]):
    def __missing__(self, key: str) -> None:
        raise ElementNotFound(key)


def init_octave_elements(pb_config: QuaConfig, octave_config: Optional[QmOctaveConfig]) -> UpconvertedElementsDB:
    elements = {}
    _octave_container = OctavesContainer(pb_config, octave_config)
    for name, element_config in pb_config.v1_beta.elements.items():
        _, element_inputs = betterproto.which_one_of(element_config, "element_inputs_one_of")
        if isinstance(element_inputs, QuaConfigMixInputs):
            input_inst = _octave_container.create_new_api_upconverted_input(element_config, name)
            if input_inst is not None:
                rf_output = _get_element_rf_output(element_config.rf_outputs, _octave_container)
                elements[name] = NewApiUpconvertedElement(
                    name=name,
                    config=element_config,
                    element_input=input_inst,
                    element_output=rf_output,
                )
    return UpconvertedElementsDB(elements)



from typing import Type

from qm.utils.async_utils import run_async
from qm.api.base_api import BaseApi, connection_error_handle
from qm.io.qualang.api.v1 import GetInfoRequest, InfoServiceStub
from qm.api.models.info import QuaMachineInfo, ImplementationInfo


@connection_error_handle()
class InfoServiceApi(BaseApi[InfoServiceStub]):
    @property
    def _stub_class(self) -> Type[InfoServiceStub]:
        return InfoServiceStub

    def get_info(self) -> QuaMachineInfo:
        request = GetInfoRequest()
        response = run_async(self._stub.get_info(request, timeout=self._timeout))

        return QuaMachineInfo(
            capabilities=response.capabilities,
            implementation=ImplementationInfo(
                name=response.implementation.name,
                version=response.implementation.version,
                url=response.implementation.url,
            ),
        )



from typing import TYPE_CHECKING, List, Tuple, Union, Optional

import betterproto
from betterproto.lib.google.protobuf import Empty

from qm._loc import _get_loc
from qm.exceptions import QmQuaException
from qm.grpc.qua import (
    QuaProgramChirp,
    QuaProgramRampPulse,
    QuaProgramCorrection,
    QuaProgramIfStatement,
    QuaProgramAnyStatement,
    QuaProgramForStatement,
    QuaProgramAmpMultiplier,
    QuaProgramPlayStatement,
    QuaProgramSaveStatement,
    QuaProgramWaitStatement,
    QuaProgramAlignStatement,
    QuaProgramMeasureProcess,
    QuaProgramPauseStatement,
    QuaProgramPulseReference,
    QuaProgramForEachStatement,
    QuaProgramMeasureStatement,
    QuaProgramVarRefExpression,
    QuaProgramZRotationStatement,
    QuaProgramAnyScalarExpression,
    QuaProgramAssignmentStatement,
    QuaProgramRampToZeroStatement,
    QuaProgramResetFrameStatement,
    QuaProgramResetPhaseStatement,
    QuaProgramSaveStatementSource,
    QuaProgramSetDcOffsetStatement,
    QuaProgramStatementsCollection,
    QuaProgramArrayVarRefExpression,
    QuaProgramStrictTimingStatement,
    QuaProgramQuantumElementReference,
    QuaProgramWaitForTriggerStatement,
    QuaProgramUpdateFrequencyStatement,
    QuaProgramAssignmentStatementTarget,
    QuaProgramResetGlobalPhaseStatement,
    QuaProgramUpdateCorrectionStatement,
    QuaProgramFastFrameRotationStatement,
    QuaProgramAdvanceInputStreamStatement,
    QuaProgramUpdateFrequencyStatementUnits,
    QuaProgramForEachStatementVariableWithValues,
    QuaProgramWaitForTriggerStatementElementOutput,
)

if TYPE_CHECKING:
    from qm.qua._dsl import _ResultSource
    from qm.qua._dsl_specific_type_hints import PlayPulseType, MeasurePulseType
    from qm.qua._type_hinting import MessageVarType, MessageExpressionType, MessageVariableOrExpression


class StatementsCollection:
    def __init__(self, body: QuaProgramStatementsCollection):
        self._body = body

    @staticmethod
    def _check_serialised_on_wire(message: QuaProgramAnyStatement, name: str) -> None:
        try:
            getattr(message, name)
        except AttributeError:
            raise QmQuaException(f"Failed to serialize of wire {name}")

    def _create_play_statement(
        self,
        pulse: "PlayPulseType",
        element: str,
        duration: Optional["MessageExpressionType"] = None,
        condition: Optional["MessageExpressionType"] = None,
        target: str = "",
        chirp: Optional[QuaProgramChirp] = None,
        truncate: Optional["MessageExpressionType"] = None,
        timestamp_label: Optional[str] = None,
    ) -> QuaProgramPlayStatement:
        amp = None
        if isinstance(pulse, tuple):
            pulse, amp = pulse

        loc = _get_loc()
        play_statement = QuaProgramPlayStatement(
            loc=loc,
            qe=QuaProgramQuantumElementReference(name=element, loc=loc),
            target_input=target,
        )
        if isinstance(pulse, QuaProgramRampPulse):
            play_statement.ramp_pulse = QuaProgramRampPulse().from_dict(pulse.to_dict())
        else:
            play_statement.named_pulse = QuaProgramPulseReference(name=pulse, loc=loc)

        if duration is not None:
            play_statement.duration = QuaProgramAnyScalarExpression().from_dict(duration.to_dict())
        if condition is not None:
            play_statement.condition = QuaProgramAnyScalarExpression().from_dict(condition.to_dict())
        if chirp is not None:
            play_statement.chirp = QuaProgramChirp().from_dict(chirp.to_dict())
            play_statement.chirp.loc = loc
        if amp is not None:
            play_statement.amp = QuaProgramAmpMultiplier(loc=loc, v0=amp[0])
            for i in range(1, 4, 1):
                if amp[i] is not None:
                    setattr(play_statement.amp, "v" + str(i), amp[i])
        if truncate is not None:
            play_statement.truncate = QuaProgramAnyScalarExpression().from_dict(truncate.to_dict())
        if timestamp_label is not None:
            play_statement.timestamp_label = timestamp_label
        return play_statement

    def play(
        self,
        pulse: "PlayPulseType",
        element: str,
        duration: Optional["MessageExpressionType"] = None,
        condition: Optional["MessageExpressionType"] = None,
        target: str = "",
        chirp: Optional[QuaProgramChirp] = None,
        truncate: Optional["MessageExpressionType"] = None,
        timestamp_label: Optional[str] = None,
    ) -> None:
        """Play a pulse to a element as per the OPX config

        Args:
            pulse: A tuple (pulse, amp). pulse is string of pulse name,
                amp is a 4 matrix
            element
            duration
            condition
            target
            chirp
            truncate
            timestamp_label

        Returns:

        """
        play_statement = self._create_play_statement(
            pulse, element, duration, condition, target, chirp, truncate, timestamp_label
        )
        statement = QuaProgramAnyStatement(play=play_statement)
        self._check_serialised_on_wire(statement, "play")
        self._body.statements.append(statement)

    def pause(self, *elements: str) -> None:
        """Pause the execution of the given elements

        Args:
            *elements

        Returns:

        """
        loc = _get_loc()
        statement = QuaProgramAnyStatement(
            pause=QuaProgramPauseStatement(
                loc=_get_loc(),
                qes=[QuaProgramQuantumElementReference(name=element, loc=loc) for element in elements],
            )
        )
        self._check_serialised_on_wire(statement, "pause")
        self._body.statements.append(statement)

    def update_frequency(
        self,
        element: str,
        new_frequency: "MessageExpressionType",
        units: str,
        keep_phase: bool,
    ) -> None:
        """Updates the frequency of a given element

        Args:
            element: The element to set the frequency to
            new_frequency: The new frequency value to set
            units
            keep_phase

        Returns:

        """
        try:
            units_enum = QuaProgramUpdateFrequencyStatementUnits[units]
        except KeyError:
            raise QmQuaException(f'unknown units "{units}"')

        loc = _get_loc()
        statement = QuaProgramAnyStatement(
            update_frequency=QuaProgramUpdateFrequencyStatement(
                loc=loc,
                qe=QuaProgramQuantumElementReference(name=element, loc=loc),
                units=units_enum,
                keep_phase=keep_phase,
                value=QuaProgramAnyScalarExpression().from_dict(new_frequency.to_dict()),
            )
        )
        self._check_serialised_on_wire(statement, "update_frequency")
        self._body.statements.append(statement)

    def update_correction(
        self,
        element: str,
        c00: "MessageExpressionType",
        c01: "MessageExpressionType",
        c10: "MessageExpressionType",
        c11: "MessageExpressionType",
    ) -> None:
        """Updates the correction of a given element

        Args:
            element: The element to set the correction to
            c00: The top left matrix element
            c01: The top right matrix element
            c10: The bottom left matrix element
            c11: The bottom right matrix element
        """
        loc = _get_loc()
        statement = QuaProgramAnyStatement(
            update_correction=QuaProgramUpdateCorrectionStatement(
                loc=loc,
                qe=QuaProgramQuantumElementReference(name=element, loc=loc),
                correction=QuaProgramCorrection(c0=c00, c1=c01, c2=c10, c3=c11),
            )
        )
        self._body.statements.append(statement)

    def set_dc_offset(self, element: str, element_input: str, offset: "MessageExpressionType") -> None:
        """Update the DC offset of an element's input

        Args:
            element: The element to update its DC offset
            element_input: desired input of the element, can be 'single'
                for a 'singleInput' element or 'I' or 'Q' for a
                'mixInputs' element
            offset: Desired dc offset for single
        """
        loc = _get_loc()
        statement = QuaProgramAnyStatement(
            set_dc_offset=QuaProgramSetDcOffsetStatement(
                loc=loc,
                qe=QuaProgramQuantumElementReference(name=element, loc=loc),
                qe_input_reference=element_input,
                offset=offset,
            )
        )
        self._check_serialised_on_wire(statement, "set_dc_offset")
        self._body.statements.append(statement)

    def advance_input_stream(self, input_stream: "MessageVariableOrExpression") -> None:
        """advance an input stream pointer to be sent to the QUA program

        Args:
            input_stream: The input stream to advance
        """
        statement = QuaProgramAnyStatement(advance_input_stream=QuaProgramAdvanceInputStreamStatement(loc=_get_loc()))

        if isinstance(input_stream, QuaProgramArrayVarRefExpression):
            statement.advance_input_stream.stream_array = input_stream
        elif (
            isinstance(input_stream, QuaProgramAnyScalarExpression)
            and betterproto.which_one_of(input_stream, "expression_oneof")[0] == "variable"
        ):
            statement.advance_input_stream.stream_variable = input_stream.variable
        else:
            raise QmQuaException("unsupported type for advance input stream")

        self._check_serialised_on_wire(statement, "advance_input_stream")
        self._body.statements.append(statement)

    def align(self, *elements: str) -> None:
        """Align the given elements

        Args:
            *elements

        Returns:

        """
        loc = _get_loc()
        statement = QuaProgramAnyStatement(
            align=QuaProgramAlignStatement(
                loc=loc,
                qe=[QuaProgramQuantumElementReference(name=element, loc=loc) for element in elements],
            )
        )
        self._check_serialised_on_wire(statement, "align")
        self._body.statements.append(statement)

    def reset_if_phase(self, element: str) -> None:
        """TODO: document

        Args:
            element

        Returns:

        """
        loc = _get_loc()
        statement = QuaProgramAnyStatement(
            reset_phase=QuaProgramResetPhaseStatement(
                loc=loc, qe=QuaProgramQuantumElementReference(name=element, loc=loc)
            )
        )
        self._check_serialised_on_wire(statement, "reset_phase")
        self._body.statements.append(statement)

    def reset_global_phase(self) -> None:
        loc = _get_loc()
        statement = QuaProgramAnyStatement(reset_global_phase=QuaProgramResetGlobalPhaseStatement(loc=loc))
        self._check_serialised_on_wire(statement, "reset_global_phase")
        self._body.statements.append(statement)

    def wait(self, duration: "MessageExpressionType", *elements: str) -> None:
        """Waits for the given duration on all provided elements

        Args:
            duration
            *elements

        Returns:

        """
        loc = _get_loc()
        statement = QuaProgramAnyStatement(
            wait=QuaProgramWaitStatement(
                loc=loc,
                time=QuaProgramAnyScalarExpression().from_dict(duration.to_dict()),
                qe=[QuaProgramQuantumElementReference(name=element, loc=loc) for element in elements],
            )
        )
        self._check_serialised_on_wire(statement, "wait")
        self._body.statements.append(statement)

    def wait_for_trigger(
        self,
        pulse_to_play: Optional[str],
        trigger_element: Optional[Union[Tuple[str, str], str]],
        time_tag_target: Optional["MessageVarType"],
        *elements: str,
    ) -> None:
        loc = _get_loc()
        statement = QuaProgramAnyStatement(
            wait_for_trigger=QuaProgramWaitForTriggerStatement(
                loc=loc,
                qe=[QuaProgramQuantumElementReference(name=element, loc=loc) for element in elements],
            )
        )

        if pulse_to_play is not None:
            statement.wait_for_trigger.pulse_to_play.name = pulse_to_play
        if trigger_element is not None:
            if isinstance(trigger_element, tuple):
                el, out = trigger_element
                statement.wait_for_trigger.element_output = QuaProgramWaitForTriggerStatementElementOutput(
                    element=el, output=out
                )
            else:
                statement.wait_for_trigger.element_output = QuaProgramWaitForTriggerStatementElementOutput(
                    element=trigger_element
                )
        else:
            statement.wait_for_trigger.global_trigger = Empty()
        if time_tag_target is not None:
            statement.wait_for_trigger.time_tag_target = time_tag_target

        self._check_serialised_on_wire(statement, "wait_for_trigger")
        self._body.statements.append(statement)

    def save(self, source: QuaProgramSaveStatementSource, result: "_ResultSource") -> None:
        statement = QuaProgramAnyStatement(
            save=QuaProgramSaveStatement(loc=_get_loc(), source=source, tag=result.get_var_name())
        )

        self._check_serialised_on_wire(statement, "save")
        self._body.statements.append(statement)

    def z_rotation(self, angle: "MessageExpressionType", *elements: str) -> None:
        loc = _get_loc()
        for element in elements:
            statement = QuaProgramAnyStatement(
                z_rotation=QuaProgramZRotationStatement(
                    loc=loc,
                    value=QuaProgramAnyScalarExpression().from_dict(angle.to_dict()),
                    qe=QuaProgramQuantumElementReference(name=element, loc=loc),
                )
            )

            self._check_serialised_on_wire(statement, "z_rotation")
            self._body.statements.append(statement)

    def reset_frame(self, *elements: str) -> None:
        loc = _get_loc()
        for element in elements:
            statement = QuaProgramAnyStatement(
                reset_frame=QuaProgramResetFrameStatement(
                    loc=loc,
                    qe=QuaProgramQuantumElementReference(name=element, loc=loc),
                )
            )

            self._check_serialised_on_wire(statement, "reset_frame")
            self._body.statements.append(statement)

    def fast_frame_rotation(
        self, cosine: QuaProgramAnyScalarExpression, sine: QuaProgramAnyScalarExpression, *elements: str
    ) -> None:
        loc = _get_loc()
        for element in elements:
            statement = QuaProgramAnyStatement(
                fast_frame_rotation=QuaProgramFastFrameRotationStatement(
                    loc=loc,
                    cosine=cosine,
                    sine=sine,
                    qe=QuaProgramQuantumElementReference(name=element, loc=loc),
                )
            )
            self._check_serialised_on_wire(statement, "fast_frame_rotation")
            self._body.statements.append(statement)

    def ramp_to_zero(self, element: str, duration: Optional[int]) -> None:
        loc = _get_loc()
        statement = QuaProgramAnyStatement(
            ramp_to_zero=QuaProgramRampToZeroStatement(
                loc=loc,
                qe=QuaProgramQuantumElementReference(name=element, loc=loc),
                duration=duration if duration else None,
            )
        )

        self._check_serialised_on_wire(statement, "ramp_to_zero")
        self._body.statements.append(statement)

    def measure(
        self,
        pulse: "MeasurePulseType",
        element: str,
        stream: Optional["_ResultSource"] = None,
        *processes: QuaProgramMeasureProcess,
        timestamp_label: Optional[str] = None,
    ) -> None:
        """Measure an element using the given pulse, process the result with the integration weights and
        store the results to the provided variables

        Args:
            pulse
            element
            stream (_ResultSource)
            *processes: an iterable of analog processes
            timestamp_label

        Returns:

        """
        amp = None
        if isinstance(pulse, tuple):
            pulse, amp = pulse

        loc = _get_loc()
        statement = QuaProgramAnyStatement(
            measure=QuaProgramMeasureStatement(
                loc=loc,
                pulse=QuaProgramPulseReference(name=pulse, loc=loc),
                qe=QuaProgramQuantumElementReference(name=element, loc=loc),
            )
        )
        if stream is not None:
            statement.measure.stream_as = stream.get_var_name()

        for analog_process in processes:
            statement.measure.measure_processes.append(analog_process)

        if amp is not None:
            statement.measure.amp.loc = loc
            statement.measure.amp.v0 = amp[0]
            for i in range(1, 4, 1):
                if amp[i] is not None:
                    setattr(statement.measure.amp, "v" + str(i), amp[i])

        if timestamp_label is not None:
            statement.measure.timestamp_label = timestamp_label

        self._check_serialised_on_wire(statement, "measure")
        self._body.statements.append(statement)

    def if_block(self, condition: "MessageExpressionType", unsafe: bool = False) -> "StatementsCollection":
        statement = QuaProgramAnyStatement(
            if_=QuaProgramIfStatement(
                loc=_get_loc(),
                condition=condition,
                unsafe=unsafe,
                body=QuaProgramStatementsCollection(statements=[]),
            )
        )

        self._check_serialised_on_wire(statement, "if_")
        self._body.statements.append(statement)
        return StatementsCollection(statement.if_.body)

    def for_each(
        self, iterators: List[Tuple[QuaProgramVarRefExpression, QuaProgramArrayVarRefExpression]]
    ) -> "StatementsCollection":
        statement = QuaProgramAnyStatement(
            for_each=QuaProgramForEachStatement(
                loc=_get_loc(),
                body=QuaProgramStatementsCollection(statements=[]),
                iterator=[
                    QuaProgramForEachStatementVariableWithValues(variable=var, array=arr) for var, arr in iterators
                ],
            )
        )

        self._check_serialised_on_wire(statement, "for_each")
        self._body.statements.append(statement)
        return StatementsCollection(statement.for_each.body)

    def get_last_statement(self) -> Optional[QuaProgramAnyStatement]:
        statements = self._body.statements
        length_statements = len(statements)
        if length_statements == 0:
            return None
        return statements[-1]

    def for_block(self) -> QuaProgramForStatement:
        statement = QuaProgramAnyStatement(
            for_=QuaProgramForStatement(loc=_get_loc(), body=QuaProgramStatementsCollection(statements=[]))
        )

        self._check_serialised_on_wire(statement, "for_")
        self._body.statements.append(statement)
        return statement.for_

    def strict_timing_block(self) -> QuaProgramStrictTimingStatement:
        statement = QuaProgramAnyStatement(
            strict_timing=QuaProgramStrictTimingStatement(
                loc=_get_loc(), body=QuaProgramStatementsCollection(statements=[])
            )
        )

        self._check_serialised_on_wire(statement, "strict_timing")
        self._body.statements.append(statement)
        return statement.strict_timing

    def assign(
        self,
        target: QuaProgramAssignmentStatementTarget,
        expression: "MessageExpressionType",
    ) -> None:
        """Assign a value calculated by :expression into :target

        Args:
            target: The name of the variable to assign to
            expression: The expression to calculate

        Returns:

        """
        statement = QuaProgramAnyStatement(
            assign=QuaProgramAssignmentStatement(loc=_get_loc(), target=target, expression=expression)
        )

        self._check_serialised_on_wire(statement, "assign")
        self._body.statements.append(statement)


class PortConditionedStatementsCollection(StatementsCollection):
    def __init__(self, body: QuaProgramStatementsCollection, condition: "MessageExpressionType"):
        super().__init__(body)
        self._port_condition = condition

    def _create_play_statement(
        self,
        pulse: "PlayPulseType",
        element: str,
        duration: Optional["MessageExpressionType"] = None,
        condition: Optional["MessageExpressionType"] = None,
        target: str = "",
        chirp: Optional[QuaProgramChirp] = None,
        truncate: Optional["MessageExpressionType"] = None,
        timestamp_label: Optional[str] = None,
    ) -> QuaProgramPlayStatement:
        play_statement = super()._create_play_statement(
            pulse, element, duration, condition, target, chirp, truncate, timestamp_label
        )
        play_statement.port_condition = QuaProgramAnyScalarExpression().from_dict(self._port_condition.to_dict())
        return play_statement



import warnings
from typing import Dict, List, Tuple, Union, Literal, cast

import betterproto

from qm.type_hinting import Number
from qm.type_hinting.config_types import (
    DictQuaConfig,
    LfFemConfigType,
    MixerConfigType,
    MwFemConfigType,
    PulseConfigType,
    StickyConfigType,
    ElementConfigType,
    MwInputConfigType,
    PortReferenceType,
    MixInputConfigType,
    MwOutputConfigType,
    ControllerConfigType,
    HoldOffsetConfigType,
    OscillatorConfigType,
    SingleInputConfigType,
    DigitalInputConfigType,
    AnalogInputPortConfigType,
    DigitalWaveformConfigType,
    InputCollectionConfigType,
    AnalogOutputPortConfigType,
    ConstantWaveFormConfigType,
    DigitalInputPortConfigType,
    ArbitraryWaveFormConfigType,
    DigitalOutputPortConfigType,
    IntegrationWeightConfigType,
    OPX1000ControllerConfigType,
    CompressedWaveFormConfigType,
    MwFemAnalogInputPortConfigType,
    MwFemAnalogOutputPortConfigType,
    AnalogOutputPortConfigTypeOctoDac,
)
from qm.grpc.qua_config import (
    QuaConfig,
    QuaConfigMatrix,
    QuaConfigSticky,
    QuaConfigFemTypes,
    QuaConfigMixerDec,
    QuaConfigPulseDec,
    QuaConfigDeviceDec,
    QuaConfigMixInputs,
    QuaConfigElementDec,
    QuaConfigHoldOffset,
    QuaConfigOscillator,
    QuaConfigQuaConfigV1,
    QuaConfigSingleInput,
    QuaConfigWaveformDec,
    QuaConfigControllerDec,
    QuaConfigElementThread,
    QuaConfigOctoDacFemDec,
    QuaConfigPortReference,
    QuaConfigMultipleInputs,
    QuaConfigMicrowaveFemDec,
    QuaConfigMultipleOutputs,
    QuaConfigAdcPortReference,
    QuaConfigDacPortReference,
    QuaConfigPulseDecOperation,
    QuaConfigAnalogInputPortDec,
    QuaConfigDigitalWaveformDec,
    QuaConfigAnalogOutputPortDec,
    QuaConfigConstantWaveformDec,
    QuaConfigDigitalInputPortDec,
    QuaConfigArbitraryWaveformDec,
    QuaConfigDigitalOutputPortDec,
    QuaConfigIntegrationWeightDec,
    QuaConfigCompressedWaveformDec,
    QuaConfigSingleInputCollection,
    QuaConfigDigitalInputPortReference,
    QuaConfigDigitalOutputPortReference,
    QuaConfigOctoDacAnalogOutputPortDec,
    QuaConfigDigitalInputPortDecPolarity,
    QuaConfigMicrowaveAnalogInputPortDec,
    QuaConfigMicrowaveInputPortReference,
    QuaConfigMicrowaveAnalogOutputPortDec,
    QuaConfigMicrowaveOutputPortReference,
    QuaConfigOctoDacAnalogOutputPortDecOutputMode,
    QuaConfigOctoDacAnalogOutputPortDecSamplingRateMode,
)


def convert_msg_to_config(config: QuaConfig) -> DictQuaConfig:
    return _convert_v1_beta(config.v1_beta)


def _convert_mixers(mixers: Dict[str, QuaConfigMixerDec]) -> Dict[str, List[MixerConfigType]]:
    ret: Dict[str, List[MixerConfigType]] = {}
    for name, data in mixers.items():
        temp_list: List[MixerConfigType] = []
        for correction in data.correction:
            if correction.frequency_double:
                frequency = correction.frequency_double
            else:
                frequency = correction.frequency

            if correction.frequency_negative:
                frequency = -frequency

            if correction.lo_frequency_double:
                lo_frequency = correction.lo_frequency_double
            else:
                lo_frequency = correction.lo_frequency

            temp_dict: MixerConfigType = {
                "intermediate_frequency": frequency,
                "lo_frequency": lo_frequency,
                "correction": _convert_matrix(correction.correction),
            }
            temp_list.append(temp_dict)

        ret[name] = temp_list
    return ret


def _convert_matrix(matrix: QuaConfigMatrix) -> Tuple[Number, Number, Number, Number]:
    return matrix.v00, matrix.v01, matrix.v10, matrix.v11


def _convert_integration_weights(
    integration_weights: Dict[str, QuaConfigIntegrationWeightDec]
) -> Dict[str, IntegrationWeightConfigType]:
    ret: Dict[str, IntegrationWeightConfigType] = {}
    for name, data in integration_weights.items():
        tmp: IntegrationWeightConfigType = {
            "cosine": [(s.value, s.length) for s in data.cosine],
            "sine": [(s.value, s.length) for s in data.sine],
        }
        ret[name] = tmp
    return ret


def _convert_digital_wave_forms(
    digital_wave_forms: Dict[str, QuaConfigDigitalWaveformDec]
) -> Dict[str, DigitalWaveformConfigType]:
    ret: Dict[str, DigitalWaveformConfigType] = {}
    for name, data in digital_wave_forms.items():
        temp_list: List[Tuple[int, int]] = []
        for sample in data.samples:
            value = int(bool(sample.value))
            temp_list.append((value, sample.length))

        ret[name] = {"samples": temp_list}
    return ret


def _convert_wave_forms(
    wave_forms: Dict[str, QuaConfigWaveformDec]
) -> Dict[str, Union[ArbitraryWaveFormConfigType, ConstantWaveFormConfigType, CompressedWaveFormConfigType]]:
    ret: Dict[str, Union[ArbitraryWaveFormConfigType, ConstantWaveFormConfigType, CompressedWaveFormConfigType]] = {}
    for name, data in wave_forms.items():
        key_name, curr_waveform = betterproto.which_one_of(data, "waveform_oneof")
        if isinstance(curr_waveform, QuaConfigArbitraryWaveformDec):

            arbitrary_waveform_dict: ArbitraryWaveFormConfigType = {
                "type": "arbitrary",
                "samples": curr_waveform.samples,
                "is_overridable": curr_waveform.is_overridable,
            }
            if isinstance(curr_waveform.max_allowed_error, float):
                arbitrary_waveform_dict["max_allowed_error"] = curr_waveform.max_allowed_error
            if isinstance(curr_waveform.sampling_rate, float):
                arbitrary_waveform_dict["sampling_rate"] = curr_waveform.sampling_rate
            ret[name] = arbitrary_waveform_dict

        elif isinstance(curr_waveform, QuaConfigConstantWaveformDec):
            constant_waveform_dict: ConstantWaveFormConfigType = {
                "type": "constant",
                "sample": curr_waveform.sample,
            }
            ret[name] = constant_waveform_dict
        elif isinstance(curr_waveform, QuaConfigCompressedWaveformDec):
            warnings.warn("Compressed waveform is deprecated.", DeprecationWarning)
            compressed_waveform_dict: CompressedWaveFormConfigType = {
                "type": "compressed",
                "samples": curr_waveform.samples,
                "sample_rate": curr_waveform.sample_rate,
            }
            ret[name] = compressed_waveform_dict
        else:
            raise Exception(f"Unknown waveform type - {key_name}")

    return ret


def _convert_pulses(pulses: Dict[str, QuaConfigPulseDec]) -> Dict[str, PulseConfigType]:
    ret = {}
    for name, data in pulses.items():
        temp_dict: PulseConfigType = {
            "length": data.length,
            "waveforms": data.waveforms,
            "integration_weights": data.integration_weights,
            "operation": cast(
                Literal["measurement", "control"], QuaConfigPulseDecOperation(data.operation).name.lower()
            ),
        }
        if isinstance(data.digital_marker, str):
            temp_dict["digital_marker"] = data.digital_marker
        ret[name] = temp_dict
    return ret


def _convert_v1_beta(config: QuaConfigQuaConfigV1) -> DictQuaConfig:
    if config.control_devices:
        controllers = _convert_controller_types(config.control_devices)
    elif config.controllers:
        controllers = {name: _convert_controller(controller) for name, controller in config.controllers.items()}
    else:
        controllers = {}
    result: DictQuaConfig = {
        "version": 1,
        "controllers": controllers,
        "oscillators": _convert_oscillators(config.oscillators),
        "elements": _convert_elements(config.elements),
        "pulses": _convert_pulses(config.pulses),
        "waveforms": _convert_wave_forms(config.waveforms),
        "digital_waveforms": _convert_digital_wave_forms(config.digital_waveforms),
        "integration_weights": _convert_integration_weights(config.integration_weights),
        "mixers": _convert_mixers(config.mixers),
    }
    return result


def _convert_controller(data: QuaConfigControllerDec) -> ControllerConfigType:
    return {
        "type": cast(Literal["opx", "opx1"], data.type),
        "analog_outputs": _convert_controller_analog_outputs(data.analog_outputs),
        "analog_inputs": _convert_controller_analog_inputs(data.analog_inputs),
        "digital_outputs": _convert_controller_digital_outputs(data.digital_outputs),
        "digital_inputs": _convert_controller_digital_inputs(data.digital_inputs),
    }


def _convert_controller_types(
    controllers: Dict[str, QuaConfigDeviceDec]
) -> Dict[str, Union[ControllerConfigType, OPX1000ControllerConfigType]]:
    ret: Dict[str, Union[ControllerConfigType, OPX1000ControllerConfigType]] = {}
    for name, data in controllers.items():
        if len(data.fems) == 1 and 1 in data.fems:
            _, opx = betterproto.which_one_of(data.fems[1], "fem_type_one_of")
            if isinstance(opx, QuaConfigControllerDec):
                ret[name] = _convert_controller(opx)
                continue

        to_attach: OPX1000ControllerConfigType = {
            "type": "opx1000",
            "fems": {fem_idx: _convert_fem(fem) for fem_idx, fem in data.fems.items()},
        }
        ret[name] = to_attach

    return ret


def _convert_fem(data: QuaConfigFemTypes) -> Union[LfFemConfigType, MwFemConfigType]:
    _, fem_config = betterproto.which_one_of(data, "fem_type_one_of")
    if isinstance(fem_config, QuaConfigOctoDacFemDec):
        return _convert_octo_dac(fem_config)
    elif isinstance(fem_config, QuaConfigMicrowaveFemDec):
        return _convert_mw_fem(fem_config)
    else:
        raise ValueError(f"Unknown FEM type - {fem_config}")


def _convert_mw_fem(data: QuaConfigMicrowaveFemDec) -> MwFemConfigType:
    ret: MwFemConfigType = {"type": "MW"}
    if data.analog_outputs:
        ret["analog_outputs"] = _convert_mw_analog_outputs(data.analog_outputs)
    if data.analog_inputs:
        ret["analog_inputs"] = _convert_mw_analog_inputs(data.analog_inputs)
    if data.digital_outputs:
        ret["digital_outputs"] = _convert_controller_digital_outputs(data.digital_outputs)
    if data.digital_inputs:
        ret["digital_inputs"] = _convert_controller_digital_inputs(data.digital_inputs)
    return ret


def _convert_mw_analog_outputs(
    outputs: Dict[int, QuaConfigMicrowaveAnalogOutputPortDec]
) -> Dict[int, MwFemAnalogOutputPortConfigType]:
    return {
        idx: {
            "sampling_rate": output.sampling_rate,
            "full_scale_power_dbm": output.full_scale_power_dbm,
            "band": cast(Literal[1, 2, 3], output.band),
            "delay": output.delay,
            "shareable": output.shareable,
            "upconverters": {k: {"frequency": v.frequency} for k, v in output.upconverters.items()},
        }
        for idx, output in outputs.items()
    }


def _convert_mw_analog_inputs(
    inputs: Dict[int, QuaConfigMicrowaveAnalogInputPortDec]
) -> Dict[int, MwFemAnalogInputPortConfigType]:
    return {
        idx: {
            "band": cast(Literal[1, 3, 3], _input.band),
            "shareable": _input.shareable,
            "gain_db": _input.gain_db,
            "sampling_rate": _input.sampling_rate,
            "downconverter_frequency": _input.downconverter.frequency,
        }
        for idx, _input in inputs.items()
    }


def _convert_octo_dac(data: QuaConfigOctoDacFemDec) -> LfFemConfigType:
    ret: LfFemConfigType = {"type": "LF"}
    if data.analog_outputs:
        ret["analog_outputs"] = _convert_octo_dac_fem_analog_outputs(data.analog_outputs)
    if data.analog_inputs:
        ret["analog_inputs"] = _convert_controller_analog_inputs(data.analog_inputs)
    if data.digital_outputs:
        ret["digital_outputs"] = _convert_controller_digital_outputs(data.digital_outputs)
    if data.digital_inputs:
        ret["digital_inputs"] = _convert_controller_digital_inputs(data.digital_inputs)
    return ret


def _convert_inputs(inputs: Dict[str, QuaConfigDigitalInputPortReference]) -> Dict[str, DigitalInputConfigType]:
    ret: Dict[str, DigitalInputConfigType] = {}
    for name, data in inputs.items():
        ret[name] = {"delay": data.delay, "buffer": data.buffer, "port": _port_reference(data.port)}
    return ret


def _convert_digital_output(outputs: Dict[str, QuaConfigDigitalOutputPortReference]) -> Dict[str, PortReferenceType]:
    ret = {}
    for name, data in outputs.items():
        ret[name] = _port_reference(data.port)

    return ret


def _convert_single_input_collection(data: QuaConfigSingleInputCollection) -> InputCollectionConfigType:
    temp = {}
    for name, input_info in data.inputs.items():
        temp[name] = _port_reference(input_info)

    res: InputCollectionConfigType = {"inputs": temp}
    return res


def _convert_multiple_inputs(data: QuaConfigMultipleInputs) -> InputCollectionConfigType:
    temp = {}
    for name, input_info in data.inputs.items():
        temp[name] = _port_reference(input_info)

    res: InputCollectionConfigType = {"inputs": temp}
    return res


def _convert_oscillators(oscillator: Dict[str, QuaConfigOscillator]) -> Dict[str, OscillatorConfigType]:
    ret: Dict[str, OscillatorConfigType] = {}
    for name, data in oscillator.items():
        oscillator_config_data: OscillatorConfigType = {}
        if data.intermediate_frequency_double:
            freq = data.intermediate_frequency_double
            oscillator_config_data["intermediate_frequency"] = freq
        elif data.intermediate_frequency:
            freq = int(data.intermediate_frequency)
            oscillator_config_data["intermediate_frequency"] = freq
        if betterproto.serialized_on_wire(data.mixer):
            if data.mixer.mixer:
                oscillator_config_data["mixer"] = data.mixer.mixer
            if data.mixer.lo_frequency_double:
                lo_freq = data.mixer.lo_frequency_double
                oscillator_config_data["lo_frequency"] = float(lo_freq)
            elif data.mixer.lo_frequency:
                lo_freq = data.mixer.lo_frequency
                oscillator_config_data["lo_frequency"] = int(lo_freq)
        ret[name] = oscillator_config_data
    return ret


def _convert_elements(elements: Dict[str, QuaConfigElementDec]) -> Dict[str, ElementConfigType]:
    ret: Dict[str, ElementConfigType] = {}
    for name, data in elements.items():
        element_config_data: ElementConfigType = {
            "digitalInputs": _convert_inputs(data.digital_inputs),
            "digitalOutputs": _convert_digital_output(data.digital_outputs),
            "outputs": _convert_element_output(
                data.outputs, betterproto.which_one_of(data, "element_outputs_one_of")[1]
            ),
            "operations": data.operations,
            "hold_offset": _convert_hold_offset(data.hold_offset),
            "sticky": _convert_sticky(data.sticky),
            "thread": _convert_element_thread(data.thread),
        }
        input_value = betterproto.which_one_of(data, "element_inputs_one_of")[1]
        if isinstance(input_value, QuaConfigSingleInput):
            element_config_data["singleInput"] = _convert_single_inputs(input_value)
        elif isinstance(input_value, QuaConfigMixInputs):
            element_config_data["mixInputs"] = _convert_mix_inputs(input_value)
        elif isinstance(input_value, QuaConfigSingleInputCollection):
            element_config_data["singleInputCollection"] = _convert_single_input_collection(input_value)
        elif isinstance(input_value, QuaConfigMultipleInputs):
            element_config_data["multipleInputs"] = _convert_multiple_inputs(input_value)
        elif isinstance(input_value, QuaConfigMicrowaveInputPortReference):
            element_config_data["MWInput"] = _convert_element_mw_input(input_value)

        output_value = betterproto.which_one_of(data, "element_outputs_one_of")[1]
        if isinstance(output_value, QuaConfigMicrowaveOutputPortReference):
            element_config_data["MWOutput"] = _convert_element_mw_output(output_value)

        if data.smearing is not None:
            element_config_data["smearing"] = data.smearing
        if data.time_of_flight is not None:
            element_config_data["time_of_flight"] = data.time_of_flight
        if data.measurement_qe:
            element_config_data["measurement_qe"] = data.measurement_qe

        oscillator = betterproto.which_one_of(data, "oscillator_one_of")[0]
        if oscillator == "named_oscillator":
            element_config_data["oscillator"] = data.named_oscillator
        else:
            sign = (-1) ** data.intermediate_frequency_negative
            if data.intermediate_frequency_double:
                element_config_data["intermediate_frequency"] = abs(data.intermediate_frequency_double) * sign
            elif data.intermediate_frequency is not None:
                element_config_data["intermediate_frequency"] = abs(data.intermediate_frequency) * sign

        ret[name] = element_config_data

    return ret


def _convert_mix_inputs(mix_inputs: QuaConfigMixInputs) -> MixInputConfigType:
    res: MixInputConfigType = {"I": _port_reference(mix_inputs.i), "Q": _port_reference(mix_inputs.q)}

    mixer = mix_inputs.mixer
    if mixer is not None:
        res["mixer"] = mixer

    if mix_inputs.lo_frequency_double:
        res["lo_frequency"] = mix_inputs.lo_frequency_double
    else:
        res["lo_frequency"] = float(mix_inputs.lo_frequency)

    return res


def _convert_single_inputs(single: QuaConfigSingleInput) -> SingleInputConfigType:
    return {"port": _port_reference(single.port)}


def _convert_hold_offset(hold_offset: QuaConfigHoldOffset) -> HoldOffsetConfigType:
    return {"duration": hold_offset.duration}


def _convert_sticky(sticky: QuaConfigSticky) -> StickyConfigType:
    res: StickyConfigType = {
        "analog": sticky.analog,
        "digital": sticky.digital,
        "duration": max(sticky.duration, 1) * 4,
    }
    return res


def _convert_element_mw_input(data: QuaConfigMicrowaveInputPortReference) -> MwInputConfigType:
    return {
        "port": _port_reference(data.port),
        "upconverter": data.upconverter,
    }


def _convert_element_mw_output(data: QuaConfigMicrowaveOutputPortReference) -> MwOutputConfigType:
    return {
        "port": _port_reference(data.port),
    }


def _convert_element_thread(element_thread: QuaConfigElementThread) -> str:
    return element_thread.thread_name


def _convert_controller_analog_outputs(
    outputs: Dict[int, QuaConfigAnalogOutputPortDec]
) -> Dict[int, AnalogOutputPortConfigType]:
    ret = {}
    for name, data in outputs.items():
        port_info: AnalogOutputPortConfigType = {
            "offset": data.offset,
            "delay": data.delay,
            "shareable": data.shareable,
            "filter": {"feedforward": data.filter.feedforward, "feedback": data.filter.feedback},
            "crosstalk": data.crosstalk,
        }
        ret[int(name)] = port_info
    return ret


def _convert_octo_dac_fem_analog_outputs(
    outputs: Dict[int, QuaConfigOctoDacAnalogOutputPortDec]
) -> Dict[int, AnalogOutputPortConfigTypeOctoDac]:
    ret = {}
    for name, data in outputs.items():
        port_info: AnalogOutputPortConfigTypeOctoDac = {
            "offset": data.offset,
            "delay": data.delay,
            "shareable": data.shareable,
            "filter": {"feedforward": data.filter.feedforward, "feedback": data.filter.feedback},
            "crosstalk": data.crosstalk,
            "output_mode": cast(
                Literal["direct", "amplified"], QuaConfigOctoDacAnalogOutputPortDecOutputMode(data.output_mode).name
            ),
        }
        if data.sampling_rate:
            port_info["sampling_rate"] = {1: 1e9, 2: 2e9}[data.sampling_rate]
        if data.upsampling_mode:
            port_info["upsampling_mode"] = cast(
                Literal["mw", "pulse"], QuaConfigOctoDacAnalogOutputPortDecSamplingRateMode(data.upsampling_mode).name
            )
        ret[int(name)] = port_info
    return ret


def _convert_controller_analog_inputs(
    inputs: Dict[int, QuaConfigAnalogInputPortDec]
) -> Dict[int, AnalogInputPortConfigType]:
    ret: Dict[int, AnalogInputPortConfigType] = {}
    for idx, data in inputs.items():
        port_info: AnalogInputPortConfigType = {
            "offset": data.offset,
            "gain_db": data.gain_db if data.gain_db is not None else 0,
            "shareable": data.shareable,
        }
        if data.sampling_rate is not None:
            port_info["sampling_rate"] = data.sampling_rate
        ret[idx] = port_info
    return ret


def _convert_controller_digital_outputs(
    outputs: Dict[int, QuaConfigDigitalOutputPortDec]
) -> Dict[int, DigitalOutputPortConfigType]:
    return {idx: _convert_controller_digital_output(data) for idx, data in outputs.items()}


def _convert_controller_digital_output(data: QuaConfigDigitalOutputPortDec) -> DigitalOutputPortConfigType:
    to_return: DigitalOutputPortConfigType = {
        "shareable": data.shareable,
        "inverted": data.inverted,
    }
    if data.level is not None:
        to_return["level"] = "TTL" if data.level == 1 else "LVTTL"
    return to_return


def _convert_controller_digital_inputs(
    inputs: Dict[int, QuaConfigDigitalInputPortDec]
) -> Dict[int, DigitalInputPortConfigType]:
    ret: Dict[int, DigitalInputPortConfigType] = {}
    for idx, data in inputs.items():
        port_info: DigitalInputPortConfigType = {
            "polarity": cast(Literal["RISING", "FALLING"], QuaConfigDigitalInputPortDecPolarity(data.polarity).name),
            "deadtime": data.deadtime,
            "threshold": data.threshold,
            "shareable": data.shareable,
        }
        ret[idx] = port_info
    return ret


def _convert_element_output(
    outputs: Dict[str, QuaConfigAdcPortReference],
    multiple_outputs: Union[QuaConfigMultipleOutputs, QuaConfigMicrowaveOutputPortReference],
) -> Dict[str, PortReferenceType]:
    if isinstance(multiple_outputs, QuaConfigMultipleOutputs):
        return {name: _port_reference(data) for name, data in multiple_outputs.port_references.items()}
    return {name: _port_reference(data) for name, data in outputs.items()}


def _port_reference(
    data: Union[QuaConfigAdcPortReference, QuaConfigDacPortReference, QuaConfigPortReference]
) -> PortReferenceType:
    if data.fem:
        return data.controller, data.fem, data.number
    else:
        return data.controller, data.number



from typing import Optional

from dependency_injector import providers, containers

from qm.api.models.info import QuaMachineInfo
from qm.api.models.capabilities import ServerCapabilities


class CapabilitiesContainer(containers.DeclarativeContainer):
    config = providers.Configuration()

    capabilities = providers.Singleton(ServerCapabilities.build, qua_implementation=config.qua_implementation)


def create_capabilities_container(
    qua_implementation: Optional[QuaMachineInfo],
) -> CapabilitiesContainer:
    container = CapabilitiesContainer()
    container.config.qua_implementation.from_value(qua_implementation)
    container.wire(
        modules=[
            "qm.program._qua_config_schema",
            "qm.program._qua_config_to_pb",
            "qm.api.job_manager_api",
            "qm.simulate.interface",
        ],
        packages=["qm.elements"],
    )
    return container



import ssl
import json
import logging
import warnings
from dataclasses import dataclass
from typing import Any, Dict, List, Union, Mapping, Iterable, Optional, TypedDict

import marshmallow
from octave_sdk.octave import OctaveDetails

from qm.api.v2.qm_api import QmApi
from qm.user_config import UserConfig
from qm.grpc.qua_config import QuaConfig
from qm.utils import deprecation_message
from qm.api.frontend_api import FrontendApi
from qm.program import Program, load_config
from qm.utils.general_utils import is_debug
from qm.type_hinting.general import PathLike
from qm.api.v2.job_api.job_api import JobData
from qm.quantum_machine import QuantumMachine
from qm.results import StreamingResultFetcher
from qm.api.models.debug_data import DebugData
from qm.jobs.simulated_job import SimulatedJob
from qm.logging_utils import set_logging_level
from qm.api.v2.job_api import JobApi, JobStatus
from qm.api.server_detector import detect_server
from qm.simulate.interface import SimulationConfig
from qm.api.job_result_api import JobResultServiceApi
from qm.persistence import BaseStore, SimpleFileStore
from qm.api.models.server_details import ServerDetails
from qm.type_hinting.config_types import DictQuaConfig
from qm.program._qua_config_to_pb import load_config_pb
from qm.octave import QmOctaveConfig, AbstractCalibrationDB
from qm.api.v2.job_api.simulated_job_api import SimulatedJobApi
from qm._octaves_container import load_config_from_calibration_db
from qm.program._qua_config_schema import validate_config_capabilities
from qm.api.simulation_api import SimulationApi, create_simulation_request
from qm.containers.capabilities_container import create_capabilities_container
from qm.octave.octave_manager import OctaveManager, prep_config_for_calibration
from qm.api.v2.qmm_api import Controller, ControllerBase, QmmApiWithDeprecations
from qm.exceptions import QmmException, ConfigSchemaError, ConfigValidationException
from qm.api.models.compiler import CompilerOptionArguments, standardize_compiler_params

logger = logging.getLogger(__name__)

Version = TypedDict(
    "Version",
    {"qm-qua": str, "QOP": str, "OPX": str, "client": str, "server": str},
    total=False,
)


@dataclass
class DevicesVersion:
    gateway: str
    controllers: Dict[str, str]
    qm_qua: str
    octaves: Dict[str, str]

    @property
    def QOP(self) -> Optional[str]:
        return SERVER_TO_QOP_VERSION_MAP.get(self.gateway)


SERVER_TO_QOP_VERSION_MAP = {
    "2.40-144e7bb": "2.0.0",
    "2.40-82d4afc": "2.0.1",
    "2.40-8521884": "2.0.2",
    "2.40-3a0d7f1": "2.0.3",
    "2.50-1a24163": "2.1.3",
    "2.60-5ba458f": "2.2.0",
    "2.60-b62e6b6": "2.2.1",
    "2.60-0b17cac": "2.2.2",
    "2.70-7abf0e0": "2.4.0",
    "2.70-ed75211": "2.4.2",
    "a6f8bc5": "3.1.0",
    "fcdfb69": "3.1.1",
    "3.0-beta-78b5e00": "3.2.0",
}


@dataclass
class Devices:
    controllers: Mapping[str, ControllerBase]
    octaves: Dict[str, OctaveDetails]


class QuantumMachinesManager:
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        *,
        cluster_name: Optional[str] = None,
        timeout: Optional[float] = None,
        log_level: int = logging.INFO,
        connection_headers: Optional[Dict[str, str]] = None,
        add_debug_data: bool = False,
        credentials: Optional[ssl.SSLContext] = None,
        store: Optional[BaseStore] = None,
        file_store_root: str = ".",
        octave: Optional[QmOctaveConfig] = None,
        octave_calibration_db_path: Optional[Union[PathLike, AbstractCalibrationDB]] = None,
    ):
        """
        Args:
            host (string): Host where to find the QM orchestrator. If
                ``None``, local settings are used
            port: Port where to find the QM orchestrator. If None, local
                settings are used
            cluster_name (string): The name of the cluster, allows redirection between devices
            timeout (float): The timeout, in seconds, for detecting the qmm. Default is 60
            log_level(string): The logging level for the connection instance. Defaults to `INFO`. Please check `logging` for available options
            octave_calibration_db_path (PathLike): The path for storing the Octave's calibration database
        """
        set_logging_level(log_level)
        self._user_config = UserConfig.create_from_file()
        self._port = port
        host = host or self._user_config.manager_host or ""
        octave_calibration_db_path = octave_calibration_db_path or self._user_config.octave_calibration_db_path
        if host is None:
            message = "Failed to connect to QuantumMachines server. No host given."
            logger.error(message)
            raise QmmException(message)

        self._cluster_name = cluster_name
        self._store = store if store else SimpleFileStore(file_store_root)
        self._server_details = self._initialize_connection(
            host=host,
            port=port,
            timeout=timeout,
            add_debug_data=add_debug_data,
            connection_headers=connection_headers,
            credentials=credentials,
        )
        if self._server_details.octaves:
            if octave is None:
                octave = QmOctaveConfig()
            if octave.get_devices():
                warnings.warn(
                    "QMM was opened with OctaveConfig. Please note that from QOP2.4.0 the octave devices "
                    "are managed by the cluster setting in the QM-app. It is recommended to remove the "
                    "OctaveConfig from the QMM instantiation.",
                    category=DeprecationWarning,
                )
            else:
                for name, info in self._server_details.octaves.items():
                    octave.add_device(name, info)
        elif octave is not None:
            opx_headers = self._server_details.connection_details.headers
            for _, device in octave.get_devices().items():
                device.headers = {**opx_headers, **device.headers}

        self._caps = self._server_details.capabilities
        self._frontend = FrontendApi(self._server_details.connection_details)
        self._simulation_api = SimulationApi(self._server_details.connection_details)
        self._octave_config = octave
        self._octave_manager = OctaveManager(self._octave_config, self, self._caps)
        self._api = None
        if self._caps.supports_api_v2:
            self._api = QmmApiWithDeprecations(
                self._server_details.connection_details,
                store=self._store,
                capabilities=self._caps,
                octave_config=self._octave_config,
                octave_manager=self._octave_manager,
            )

        raise_on_error = self._user_config.strict_healthcheck is not False
        self.perform_healthcheck(raise_on_error)
        if octave_calibration_db_path is not None and self._octave_config:
            if self._octave_config.calibration_db is not None:
                raise QmmException(
                    "Duplicate calibration_db path detected, please set the calibration db only through the QMM."
                )
            else:
                self._octave_config.set_calibration_db_without_warning(octave_calibration_db_path)

    def _initialize_connection(
        self,
        host: str,
        port: Optional[int],
        timeout: Optional[float],
        add_debug_data: bool,
        credentials: Optional[ssl.SSLContext],
        connection_headers: Optional[Dict[str, str]],
    ) -> ServerDetails:
        server_details = detect_server(
            cluster_name=self._cluster_name,
            user_token=self._user_config.user_token,
            ssl_context=credentials,
            host=host,
            port_from_user_config=self._user_config.manager_port,
            user_provided_port=port,
            add_debug_data=add_debug_data,
            timeout=timeout,
            extra_headers=connection_headers,
        )
        create_capabilities_container(server_details.qua_implementation)
        return server_details

    @property
    def store(self) -> BaseStore:
        return self._store

    @property
    def octave_manager(self) -> OctaveManager:
        # warnings.warn(
        #     "Do not use OctaveManager, it will be removed in the next version", DeprecationWarning, stacklevel=2
        # )
        return self._octave_manager

    @property
    def cluster_name(self) -> str:
        return self._cluster_name or "any"

    def perform_healthcheck(self, strict: bool = True) -> None:
        """Perform a health check against the QM server.

        Args:
            strict: Will raise an exception if health check failed
        """
        if self._api:
            self._api.perform_healthcheck()
        else:
            self._frontend.healthcheck(strict)
        if self._octave_config is not None:
            for octave in self._octave_config.get_devices():
                octave_client = self._octave_manager.get_client(octave)
                octave_client.perform_healthcheck()

    def version_dict(self) -> Version:
        """
        Returns:
            A dictionary with the qm-qua and QOP versions
        """
        from qm.version import __version__

        output_dict: Version = {}
        server_version = self._server_details.server_version
        output_dict["qm-qua"] = __version__
        if server_version in SERVER_TO_QOP_VERSION_MAP:
            output_dict["QOP"] = SERVER_TO_QOP_VERSION_MAP[server_version]
        else:
            output_dict["OPX"] = server_version
        if is_debug():
            logger.debug(f"OPX version: {server_version}")

        return output_dict

    def version(self) -> Union[Version, DevicesVersion]:
        """
        Returns:
            An object with the qm-qua and QOP versions
        """
        from qm.version import __version__

        octaves = {}
        if self._octave_config is not None:
            for octave_name in self._octave_config.get_devices():
                octaves[octave_name] = self._octave_manager.get_client(octave_name).get_version()

        if self._api is None:
            gateway = self._server_details.server_version
            controllers = self._get_controllers_as_dict()
        else:
            response = self._api.get_version()
            gateway = response.gateway
            controllers = response.controllers
        return DevicesVersion(
            gateway=gateway,
            controllers={k: None for k in controllers},
            qm_qua=__version__,
            octaves=octaves,
        )

    def reset_data_processing(self) -> None:
        """Stops current data processing for ALL running jobs"""
        if self._api:
            self._api.reset_data_processing()
        else:
            self._frontend.reset_data_processing()

    def open_qm(
        self,
        config: DictQuaConfig,
        close_other_machines: bool = True,
        validate_with_protobuf: bool = False,
        add_calibration_elements_to_config: bool = True,
        use_calibration_data: bool = True,
        keep_dc_offsets_when_closing: bool = False,
        **kwargs: Any,
    ) -> Union[QuantumMachine, QmApi]:
        """Opens a new quantum machine. A quantum machine can use multiple OPXes, and a
        single OPX can also be used by multiple quantum machines as long as they do not
        share the same physical resources (input/output ports) as defined in the config.

        Args:
            config: The config that will be used by the quantum machine
            close_other_machines: When set to true (default) any open
                quantum machines will be closed. This simplifies the
                workflow, but does not enable opening more than one
                quantum machine.
            validate_with_protobuf (bool): Validates config with
                protobuf instead of marshmallow. It is usually faster
                when working with large configs. Defaults to False.
            add_calibration_elements_to_config: Automatically adds config entries to allow Octave calibration.
            use_calibration_data: Automatically load updated calibration data from calibration database into the config.
            keep_dc_offsets_when_closing: Available in QOP 2.4.2 - When closing the QM, do not change the DC offsets.

        Returns:
            A quantum machine obj that can be used to execute programs
        """
        if kwargs:
            logger.warning(f"unused kwargs: {list(kwargs)}, please remove them.")

        loaded_config = self._load_config(config, disable_marshmallow_validation=validate_with_protobuf)

        if use_calibration_data and self._octave_config is not None and self._octave_config.devices:
            if self._octave_config.calibration_db is not None:
                loaded_config = load_config_from_calibration_db(
                    loaded_config, self._octave_config.calibration_db, self._octave_config, self._caps
                )
            else:
                logger.warning("No calibration_db set in octave config, skipping loading calibration data")

        if add_calibration_elements_to_config and self._octave_config is not None and self._octave_config.devices:
            loaded_config = prep_config_for_calibration(loaded_config, self._octave_config, self._caps)

        self._octave_manager.set_octaves_from_qua_config(loaded_config.v1_beta.octaves)
        if self._api:
            if keep_dc_offsets_when_closing:
                raise NotImplementedError("keep_dc_offsets_when_closing is not supported in the current QOP version")
            return self._api.open_qm(loaded_config, close_other_machines)

        if not self._caps.supports_keeping_dc_offsets:
            if keep_dc_offsets_when_closing:
                raise QmmException(
                    "The server does not support keeping DC offsets when closing. "
                    "Please upgrade the server to a version that supports this feature."
                )
        machine_id = self._frontend.open_qm(
            loaded_config, close_other_machines, keep_dc_offsets_when_closing=keep_dc_offsets_when_closing
        )

        return QuantumMachine(
            machine_id=machine_id,
            pb_config=loaded_config,
            frontend_api=self._frontend,
            capabilities=self._caps,
            store=self.store,
            octave_config=self._octave_config,
            octave_manager=self._octave_manager,
        )

    def _load_config(self, qua_config: DictQuaConfig, *, disable_marshmallow_validation: bool = False) -> QuaConfig:
        try:
            if disable_marshmallow_validation:
                loaded_config = load_config_pb(qua_config)
            else:
                loaded_config = load_config(qua_config)
            validate_config_capabilities(loaded_config, self._caps)
            return loaded_config
        except KeyError as key_error:
            raise ConfigValidationException(f"Missing key {key_error} in config") from key_error
        except marshmallow.exceptions.ValidationError as validation_error:
            raise ConfigSchemaError(validation_error) from validation_error

    def validate_qua_config(self, qua_config: DictQuaConfig) -> None:
        """
        Validates a qua config based on the connected server's capabilities.
        Raises an exception if the config is invalid.

        Args:
            qua_config: A python dict containing the qua config to validate
        """
        self._load_config(qua_config)

    def open_qm_from_file(self, filename: str, close_other_machines: bool = True) -> Union[QuantumMachine, QmApi]:
        """Opens a new quantum machine with config taken from a file on the local file system

        Args:
            filename: The path to the file that contains the config
            close_other_machines: Flag whether to close all other
                running machines

        Returns:
            A quantum machine obj that can be used to execute programs
        """
        warnings.warn(
            deprecation_message(
                "qmm.open_qm_from_file",
                "1.2.0",
                "1.3.0",
                "This method is going to be removed.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        with open(filename) as json_file:
            json1_str = json_file.read()

            def remove_nulls(d: Dict[Any, Any]) -> Dict[Any, Any]:
                return {k: v for k, v in d.items() if v is not None}

            config = json.loads(json1_str, object_hook=remove_nulls)
        return self.open_qm(config, close_other_machines)

    def simulate(
        self,
        config: DictQuaConfig,
        program: Program,
        simulate: SimulationConfig,
        compiler_options: Optional[CompilerOptionArguments] = None,
        *,
        strict: Optional[bool] = None,
        flags: Optional[List[str]] = None,
    ) -> Union[SimulatedJob, SimulatedJobApi]:
        """Simulate the outputs of a deterministic QUA program.

        The following example shows a simple execution of the simulator, where the
        associated config object is omitted for brevity.

        Example:
            ```python
            from qm.qua import *
            from qm import SimulationConfig, QuantumMachinesManager

            qmm = QuantumMachinesManager()

            with program() as prog:
                play('pulse1', 'qe1')

            job = qmm.simulate(config, prog, SimulationConfig(duration=100))
            ```
        Args:
            config: A QM config
            program: A QUA ``program()`` object to execute
            simulate: A ``SimulationConfig`` configuration object
            compiler_options: additional parameters to pass to execute
            strict: a deprecated option for the compiler
            flags: deprecated way to provide flags to the compiler

        Returns:
            a ``QmJob`` object (see QM Job API).
        """
        standardized_options = standardize_compiler_params(compiler_options, strict, flags)
        pb_config = load_config(config)

        if self._api:
            request = create_simulation_request(pb_config, program, simulate, standardized_options)
            simulated_job = self._api.simulate(
                request.config, request.high_level_program, request.simulate, request.controller_connections
            )
            return simulated_job
        else:
            job_id, simulated_response_part = self._simulation_api.simulate(
                pb_config, program, simulate, standardized_options
            )
            return SimulatedJob(
                job_id=job_id,
                frontend_api=self._frontend,
                capabilities=self._server_details.capabilities,
                store=self.store,
                simulated_response=simulated_response_part,
            )

    def list_open_qms(self) -> List[str]:
        """Return a list of open quantum machines. (Returns only the ids, use ``get_qm(...)`` to get the machine object)

        Returns:
            The ids list
        """
        if self._api:
            return self._api.list_open_qms()
        else:
            return self._frontend.list_open_quantum_machines()

    def list_open_quantum_machines(self) -> List[str]:
        warnings.warn(
            deprecation_message(
                "qmm.list_open_quantum_machines",
                "1.2.0",
                "1.4.0",
                "This method was renamed to `qmm.list_open_qms()`",
            )
        )
        return self.list_open_qms()

    def get_qm(self, machine_id: str) -> Union[QuantumMachine, QmApi]:
        """Gets an open quantum machine object with the given machine id

        Args:
            machine_id: The id of the open quantum machine to get

        Returns:
            A quantum machine obj that can be used to execute programs
        """
        if self._api:
            return self._api.get_qm(machine_id)

        qm_data = self._frontend.get_quantum_machine(machine_id)
        machine_id = qm_data.machine_id
        pb_config = qm_data.config

        return QuantumMachine(
            machine_id=machine_id,
            pb_config=pb_config,
            frontend_api=self._frontend,
            capabilities=self._caps,
            store=self.store,
            octave_manager=self.octave_manager,
        )

    def get_job_result_handles(self, job_id: str) -> StreamingResultFetcher:
        """
        -- Available in QOP 2.x --

        Returns the result handles for a job.
        Args:
            job_id: The job id
        Returns:
            The handles that this job generated
        """
        if self._api is not None:
            raise NotImplementedError(
                "This method is not available in the current QOP version, "
                "please use `qmm.get_job(job_id).result_handles`"
            )
        return StreamingResultFetcher(
            job_id,
            JobResultServiceApi(self._server_details.connection_details, job_id),
            self._store,
            self._caps,
        )

    def close_all_qms(self) -> None:
        """Closes ALL open quantum machines"""
        if self._api:
            self._api.close_all_qms()
        else:
            self._frontend.close_all_quantum_machines()

    def close_all_quantum_machines(self) -> None:
        warnings.warn(
            deprecation_message(
                "qmm.close_all_quantum_machines",
                "1.2.0",
                "1.4.0",
                "This function is going to change its name to `qmm.close_all_qms()`",
            ),
            category=DeprecationWarning,
            stacklevel=2,
        )
        self.close_all_qms()

    def get_controllers(self) -> List[ControllerBase]:
        """Returns a list of all the controllers that are available"""
        if self._api:
            warnings.warn(
                deprecation_message(
                    "get_controllers",
                    "1.2.0",
                    "1.3.0",
                    "This will have a different return type",
                ),
                category=DeprecationWarning,
                stacklevel=2,
            )
            new_response = self._api.get_controllers()
            return list(new_response.values())
        else:
            old_response = self._frontend.get_controllers()
            return [Controller(message.name, message.temperature) for message in old_response]

    def _get_controllers_as_dict(self) -> Mapping[str, ControllerBase]:
        if self._api:
            return self._api.get_controllers()
        else:
            old_response = self._frontend.get_controllers()
            return {message.name: Controller(message.name, message.temperature) for message in old_response}

    def get_devices(self) -> Devices:
        controllers = self._get_controllers_as_dict()
        octaves: Dict[str, OctaveDetails] = {}
        if self._octave_config is not None:
            for octave_name in self._octave_config.devices:
                octave_client = self._octave_manager.get_client(octave_name)
                octaves[octave_name] = octave_client.get_details()
        return Devices(controllers=controllers, octaves=octaves)

    def clear_all_job_results(self) -> None:
        """Deletes all data from all previous jobs"""
        if self._api:
            warnings.warn(
                deprecation_message(
                    "clear_all_job_results",
                    "1.2.0",
                    "1.3.0",
                    "This method is going to be removed.",
                ),
                category=DeprecationWarning,
                stacklevel=1,
            )
            self._api.clear_all_job_results()
        else:
            self._frontend.clear_all_job_results()

    def get_jobs(
        self,
        qm_ids: Iterable[str] = tuple(),
        job_ids: Iterable[str] = tuple(),
        user_ids: Iterable[str] = tuple(),
        description: str = "",
        status: Iterable[JobStatus] = tuple(),
    ) -> List[JobData]:
        """
        -- Available in QOP 3.x --

        Get jobs based on filtering criteria. All fields are optional.

        Args:
            qm_ids: A list of qm ids
            job_ids: A list of jobs ids
            user_ids: A list of user ids
            description: Jobs' description
            status: A list of job statuses
        Returns:
            A list of jobs
        """
        if self._api is None:
            raise NotImplementedError("This method is not available in the current QOP version")
        return self._api.get_jobs(
            qm_ids=qm_ids, job_ids=job_ids, user_ids=user_ids, description=description, status=status
        )

    def get_job(self, job_id: str) -> JobApi:
        """
        -- Available in QOP 3.x --

        Get a job based on the job_id.

        Args:
            job_id: A list of jobs ids
        Returns:
            The job
        """
        if self._api is None:
            raise NotImplementedError("This method is not available in the current QOP version")
        return self._api.get_job(job_id)

    @property
    def _debug_data(self) -> Optional[DebugData]:
        return self._server_details.connection_details.debug_data



import logging
from typing import Collection

from marshmallow import ValidationError

from qm.exceptions import ConfigValidationException
from qm.type_hinting.config_types import ElementConfigType

logger = logging.getLogger(__name__)


def validate_timetagging_parameters(data: ElementConfigType) -> None:
    if "outputPulseParameters" in data:
        pulse_parameters = data["outputPulseParameters"]
        needed_parameters = [
            "signalThreshold",
            "signalPolarity",
            "derivativeThreshold",
            "derivativePolarity",
        ]
        missing_parameters = []
        for parameter in needed_parameters:
            if parameter not in pulse_parameters:
                missing_parameters.append(parameter)
        if len(missing_parameters) > 0:
            raise ConfigValidationException(
                "An element defining the output pulse parameters must either "
                f"define all of the parameters: {needed_parameters}. "
                f"Parameters defined: {pulse_parameters}"
            )
        valid_polarity = {"ASCENDING", "DESCENDING", "ABOVE", "BELOW"}
        if data["outputPulseParameters"]["signalPolarity"].upper() not in valid_polarity:
            raise ConfigValidationException(
                f"'signalPolarity' is {data['outputPulseParameters']['signalPolarity'].upper()} but it must be one of {valid_polarity}"
            )
        if data["outputPulseParameters"]["derivativePolarity"].upper() not in valid_polarity:
            raise ConfigValidationException(
                f"'derivativePolarity' is {data['outputPulseParameters']['derivativePolarity'].upper()} but it must be one of {valid_polarity}"
            )


def _element_has_outputs(data: ElementConfigType) -> bool:
    return bool(data.get("outputs")) or bool(data.get("RF_outputs")) or bool(data.get("MWOutput"))


def _validate_existence_of_field(data: ElementConfigType, field_name: str) -> None:
    if _element_has_outputs(data) and field_name not in data:
        raise ValidationError(f"An element with an output must have {field_name} defined")
    if not _element_has_outputs(data) and field_name in data:
        if "outputs" in data:
            logger.warning(
                f"The field `{field_name}` exists though the element has no outputs (empty dict). "
                f"This behavior is going to cause `ValidationError` in the future."
            )
            return
        raise ValidationError(f"{field_name} should be used only with elements that have outputs")


def validate_output_tof(data: ElementConfigType) -> None:
    _validate_existence_of_field(data, "time_of_flight")


def validate_output_smearing(data: ElementConfigType) -> None:
    if _element_has_outputs(data) and "smearing" not in data:
        data["smearing"] = 0
    _validate_existence_of_field(data, "smearing")


def validate_oscillator(data: ElementConfigType) -> None:
    if "intermediate_frequency" in data and "oscillator" in data:
        raise ValidationError("'intermediate_frequency' and 'oscillator' cannot be defined together")


def validate_used_inputs(data: Collection[str]) -> None:
    used_inputs = set(data) & {"singleInput", "mixInputs", "singleInputCollection", "multipleInputs"}
    if len(used_inputs) > 1:
        raise ValidationError(
            f"Can't support more than a single input type. " f"Used {', '.join(used_inputs)}",
            field_name="",
        )


def validate_sticky_duration(duration: int) -> None:
    if (duration % 4) != 0:
        raise ValidationError(
            "Sticky's element duration must be a dividable by 4",
            field_name="duration",
        )


def validate_arbitrary_waveform(is_overridable: bool, has_max_allowed_error: bool, has_sampling_rate: bool) -> None:
    if is_overridable and has_max_allowed_error:
        raise ValidationError("Overridable waveforms cannot have property 'max_allowed_error'")
    if is_overridable and has_sampling_rate:
        raise ValidationError("Overridable waveforms cannot have property 'sampling_rate_key'")
    if has_max_allowed_error and has_sampling_rate:
        raise ValidationError("Cannot use both 'max_allowed_error' and 'sampling_rate'")



import logging
from typing import Dict, List, Type, Tuple, AsyncIterator

from qm.utils.async_utils import run_async
from qm.api.models.server_details import ConnectionDetails
from qm.api.base_api import BaseApi, connection_error_handle
from qm.StreamMetadata import StreamMetadata, StreamMetadataError, _get_stream_metadata_dict_from_proto_resp
from qm.grpc.results_analyser import (
    GetJobStateRequest,
    GetJobErrorsRequest,
    GetJobStateResponse,
    JobResultsServiceStub,
    GetJobNamedResultRequest,
    GetJobErrorsResponseError,
    GetJobNamedResultResponse,
    GetJobResultSchemaRequest,
    GetProgramMetadataRequest,
    GetJobResultSchemaResponse,
    GetJobNamedResultHeaderRequest,
    GetJobNamedResultHeaderResponse,
)

logger = logging.getLogger(__name__)


@connection_error_handle()
class JobResultServiceApi(BaseApi[JobResultsServiceStub]):
    def __init__(self, connection_details: ConnectionDetails, job_id: str):
        super().__init__(connection_details)
        self._id = job_id

    @property
    def id(self) -> str:
        return self._id

    @property
    def _stub_class(self) -> Type[JobResultsServiceStub]:
        return JobResultsServiceStub

    def get_job_errors(self) -> List[GetJobErrorsResponseError]:
        request = GetJobErrorsRequest(job_id=self._id)
        response = run_async(self._stub.get_job_errors(request, timeout=self._timeout))
        return response.errors

    def get_job_named_result(
        self, output_name: str, long_offset: int, limit: int
    ) -> AsyncIterator[GetJobNamedResultResponse]:
        request = GetJobNamedResultRequest(
            job_id=self._id, output_name=output_name, long_offset=long_offset, limit=limit
        )
        return self._stub.get_job_named_result(request, timeout=self._timeout)

    def get_job_state(self) -> GetJobStateResponse:
        request = GetJobStateRequest(job_id=self._id)
        response = run_async(self._stub.get_job_state(request, timeout=self._timeout))
        return response

    def get_named_header(self, output_name: str, flat_struct: bool) -> GetJobNamedResultHeaderResponse:
        request = GetJobNamedResultHeaderRequest(job_id=self._id, output_name=output_name, flat_format=flat_struct)
        response = run_async(self._stub.get_job_named_result_header(request, timeout=self._timeout))
        return response

    def get_program_metadata(self) -> Tuple[List[StreamMetadataError], Dict[str, StreamMetadata]]:
        request = GetProgramMetadataRequest(job_id=self._id)

        response = run_async(self._stub.get_program_metadata(request, timeout=self._timeout))

        if response.success:
            metadata_errors = [
                StreamMetadataError(error.error, error.location)
                for error in response.program_stream_metadata.stream_metadata_extraction_error
            ]
            metadata_dict = _get_stream_metadata_dict_from_proto_resp(response.program_stream_metadata)
            return metadata_errors, metadata_dict
        logger.warning(f"Failed to fetch program metadata for job: {self._id}")
        return [], {}

    def get_job_result_schema(self) -> GetJobResultSchemaResponse:
        request = GetJobResultSchemaRequest(job_id=self._id)
        response = run_async(self._stub.get_job_result_schema(request, timeout=self._timeout))
        return response
